"""
Implements the RunQueueTableModel class. Enabling the user to display a RunQueue in a QT-view (e.g. QTableView).
"""

# from MachineLearning.framework.RunQueueClient import RunQueueClient
import logging
import os
import textwrap
import traceback
import typing
from enum import IntEnum

import dill
from PySide6 import QtCore, QtGui, QtWidgets

# from MachineLearning.framework.RunQueue import RunQueue, RunQueueItem, RunQueueItemStatus, QueueItemActions
from configurun.classes.run_queue import (RunQueue,
                                          RunQueueHasRunningItemsException,
                                          RunQueueItem, RunQueueItemActions,
                                          RunQueueItemStatus)

log = logging.getLogger(__name__)



class RunQueueTableModel(QtCore.QAbstractTableModel):

	"""
	Class that resides between the RunQueue and the view (QTableView) and provides the data for the view in a
	convenient way. Since Runqueue can also interface over a network - this class can be used to somewhat
	buffer the data and increase responsiveness of the UI.
	"""

	autoProcessingStateChanged = QtCore.Signal(bool)

	itemHighlightIdChanged = QtCore.Signal(int) #Emitted when the highlighted item changes, emits the new id

	class CustomDataRoles(IntEnum):
		"""
		Enum containing custom data roles that can be used to retrieve data from the model
		"""
		#pylint: disable=invalid-name
		IDRole = QtCore.Qt.ItemDataRole.UserRole.value + 1
		ActionRole = QtCore.Qt.ItemDataRole.UserRole.value + 2 #An action is being performed on this item
		StatusRole = QtCore.Qt.ItemDataRole.UserRole.value + 3 #The status of this item (Queued, Running etc.)


	def __init__(self,
	    	run_queue : RunQueue,
			parent: typing.Optional[QtCore.QObject] = None
		) -> None:
		super().__init__(parent)

		self._run_queue_connections = [] #List of signal connectiosn to the run_queue, enables us to disconnect them
		self._run_queue = run_queue

		self._cur_queue_copy = []
		self._cur_run_queue_item_dict_copy = {}
		self._cur_item_dict_id_order = [] #The order of the ids in the current run_queue_item_dict, this determines
			# the order of the items in the model since the order of the dict itself is not neccessarily guaranteed
			# and because it makes sense to order by id because this makes it so insertions always happen at the end

		self.set_run_queue(run_queue) #Connects signals etc.



		self._cur_actions = [] #The actions possible for the currently hightlighted item


		#======================= Some icons =======================
		self._selection_pixmap = QtWidgets.QStyle.StandardPixmap.SP_DialogApplyButton #Checkmark #type: ignore
		self._selection_icon = QtWidgets.QApplication.style().standardIcon(self._selection_pixmap)

		self._waiting_pixmap = QtWidgets.QStyle.StandardPixmap.SP_FileDialogDetailedView #Pause button #type: ignore
		self._waiting_icon = QtWidgets.QApplication.style().standardIcon(self._waiting_pixmap)

		self._running_pixmap = QtWidgets.QStyle.StandardPixmap.SP_MediaPlay #Play button #type: ignore
		self._running_icon = QtWidgets.QApplication.style().standardIcon(self._running_pixmap)

		self._finished_pixmap = QtWidgets.QStyle.StandardPixmap.SP_DialogApplyButton #Checkmark #type: ignore
		self._finished_icon = QtWidgets.QApplication.style().standardIcon(self._finished_pixmap)

		self._stopped_pixmap = QtWidgets.QStyle.StandardPixmap.SP_DialogCancelButton #X #type: ignore
		self._stopped_icon = QtWidgets.QApplication.style().standardIcon(self._stopped_pixmap)

		self._cancelled_pixmap = QtWidgets.QStyle.StandardPixmap.SP_DialogNoButton #red circle #type: ignore
		self._cancelled_icon = QtWidgets.QApplication.style().standardIcon(self._cancelled_pixmap)

		self._failed_pixmap = QtWidgets.QStyle.StandardPixmap.SP_BrowserStop #X #type: ignore
		self._failed_icon = QtWidgets.QApplication.style().standardIcon(self._failed_pixmap)


		self._icon_dict = {
			RunQueueItemStatus.Queued: self._waiting_icon,
			RunQueueItemStatus.Running: self._running_icon,
			RunQueueItemStatus.Finished: self._finished_icon,
			RunQueueItemStatus.Stopped: self._stopped_icon,
			RunQueueItemStatus.Cancelled: self._cancelled_icon,
			RunQueueItemStatus.Failed: self._failed_icon
		}

		#=== Font for Highlighting ===
		self._highlight_font = QtGui.QFont()
		self._highlight_font.setBold(True)

		#======= Other =======
		self._highlighted_id = None
		self._prev_highlighted_id = None
		self._cur_autoprocessing_state = False


	def set_run_queue(self, new_run_queue : RunQueue) -> None:
		"""Sets the runqueue of this model to the given run_queue. This will disconnect all signals from the previous
		run_queue and connect the signals from the new run_queue

		Args:
			run_queue (RunQueue): The new run_queue to use
			also_instantiate (bool, optional): Whether to request the current state of the run_queue and store it in
			this model. Defaults to True. If False, the model will only be updated upon the next change to the run_queue
			This is useful when the runqueue is an instance of RunQueueClient, since it will not have any data until
			an authenticated server is connected.
		"""
		self.beginResetModel() #Invalidate all data in current views
		self._run_queue = new_run_queue

		#============= Signal connections linking UI to RunQueue =============
		for connection in self._run_queue_connections:
			connection.disconnect()

		#NOTE: runqueue connections are not qt-signals, insertions etc. are not necessarily done in the main thread
		self._run_queue_connections.append(self._run_queue.queueChanged.connect(self._queue_changed))
		self._run_queue_connections.append( #On item-insertion in runQueue
			self._run_queue.allItemsDictInsertion.connect(self._handle_run_queue_insertion)
		)
		self._run_queue_connections.append( #On item deletion in runQueue
			self._run_queue.allItemsDictRemoval.connect(self._handle_run_queue_deletion)
		)
		self._run_queue_connections.append(self._run_queue.itemDataChanged.connect(self._run_item_changed))
		self._run_queue_connections.append(self._run_queue.resetTriggered.connect(self.reset_model))
		self._run_queue_connections.append(self._run_queue.autoProcessingStateChanged.connect(
			self.autoprocessing_state_changed))

		self._reset_model()
		self.endResetModel()

	def stop_autoprocessing(self):
		"""Signal current runqueue to stop autoqueueing"""
		self._run_queue.stop_autoprocessing()

	def start_autoprocessing(self):
		"""Signal current runqueue to start autoprocessing items in the runqueue"""
		self._run_queue.start_autoprocessing()

	def autoprocessing_state_changed(self, new_state : bool) -> None:
		"""Synchronizes autoprocessing state to RunQueue
		"""
		self.autoProcessingStateChanged.emit(new_state) #Emit signal to UI
		self._cur_autoprocessing_state = new_state

	def get_running_configuration_count(self) -> int:
		"""Return the number of configs currently running

		Returns:
			int: The number of configs currently running
		"""
		return self._run_queue.get_running_configuration_count()

	def is_autoprocessing_enabled(self) -> bool:
		"""Return whether autoprocessing is enabled in the current runqueue"""
		return self._cur_autoprocessing_state

	def reset_model(self) -> None:
		"""Reset the model, this will emit the beginResetModel and endResetModel signals"""
		self.beginResetModel()
		self._reset_model()
		self.endResetModel()

	def _reset_model(self) -> None:
		"""Private variant of reset_model, does not emit the beginResetModel and endResetModel signals
		"""
		if self._run_queue is not None:
			self._cur_queue_copy = self._run_queue.get_queue_snapshot_copy()
			self._cur_run_queue_item_dict_copy = self._run_queue.get_all_items_dict_snapshot_copy()
			self._cur_autoprocessing_state = self._run_queue.is_autoprocessing_enabled()
			if self._cur_queue_copy is None:
				self._cur_queue_copy = []
			if self._cur_run_queue_item_dict_copy is None:
				self._cur_run_queue_item_dict_copy = {}
			self._cur_item_dict_id_order = list(self._cur_run_queue_item_dict_copy.keys())
			self._cur_item_dict_id_order.sort()



	column_names = { #Used to map column index to a name/property of a RunQueueItem
		0: ("name", "Name"),
		1: ("status", "Status"),
		2: ("item_id", "ID"),
		3: ("dt_added", "Added"),
		4: ("dt_started", "Started"),
		5: ("config", "Config"),
		6: ("dt_done", "Finished"),
		7: ("exit_code", "Exit Code"),
		8: ("stderr", "Stderr")
	}
	column_positions = { #Opposite of column_names
		"name": 0,
		"status": 1,
		"item_id": 2,
		"dt_added": 3,
		"dt_started": 4,
		"config": 5,
		"dt_done": 6,
		"exit_code": 7,
		"stderr": 8
	}

	def load_from_file(self,
		    file_path : str,
			allow_load_running_items : typing.Literal["allow", "ask", "disallow"] = "ask"
		):
		"""Loads the runqueue from a file

		Args:
			path (str): The path to the file to load from
			allow_load_running_items (typing.Literal["allow", "ask", "disallow"], optional): Whether to allow loading
				a RunQueue in which 1 or more items were running at save-time. Defaults to "ask".
		"""

		#Check if user cancelled
		if file_path is None or file_path == "":
			log.info("User cancelled loading queue from file.")
			return
		try:
			try:
				# self._run_queue.load_queue_contents_from_file(file_path)
				load_dict = RunQueue.get_queue_contents_dict_from_file(file_path)
			except RunQueueHasRunningItemsException as exception:
				if allow_load_running_items == "disallow":
					return
				elif allow_load_running_items == "allow":
					pass
				elif allow_load_running_items == "ask":
					msg = f"{type(exception).__name__}: {exception}"
					log.warning(f"Could not load RunQueue from file - {msg}")
					msg_box = QtWidgets.QMessageBox()
					msg_box.setIcon(QtWidgets.QMessageBox.Icon.Warning)
					msg_box.setWindowTitle("Could not load RunQueue from file")
					msg_box.setText("<b>Do you want to set load mode to allow importing configurations that were saved when "
						"1 or more configurations were running?</b>")
					msg_box.setInformativeText(f"<b>{type(exception).__name__}:</b> {exception}<br><br>All running items "
						"of this backup were saved with a 'stopped'-state, this might indicate that the loaded runQueue-data "
						"is not entirely up to date."
						)
					msg_box.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
					msg_box.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
					msg_box.setWindowFlags(QtCore.Qt.WindowType.WindowStaysOnTopHint)
					msg_box.show()
					msg_box.activateWindow()
					msg_box.exec()
					if msg_box.result() == QtWidgets.QMessageBox.StandardButton.Yes:
						load_dict = RunQueue.get_queue_contents_dict_from_file(file_path, allow_load_running_items=True)
					else:
						log.info("User cancelled loading queue from file.")
						return

			self._run_queue.load_queue_contents_dict(load_dict) #type: ignore #Actually load the queue-data into existing

		except Exception as exception: #pylint: disable=broad-exception-caught
			msg = f"{type(exception).__name__}: {exception}"
			trace = traceback.format_exc()
			log.warning(f"Could not load RunQueue from path {file_path} - {msg}")
			msg_box = QtWidgets.QMessageBox()
			msg_box.setIcon(QtWidgets.QMessageBox.Icon.Critical)
			msg_box.setWindowTitle("Could not load RunQueue from file")
			msg_box.setText("Could not load RunQueue from file due to an unexpected error.")
			msg_box.setInformativeText(msg)
			msg_box.setDetailedText(trace)
			log.warning(trace)
			msg_box.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
			msg_box.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Ok)
			msg_box.show()
			msg_box.activateWindow()
			msg_box.exec()
			return

		log.info(f"Loaded RunQueue from file {file_path}.")

	def load_from_file_popup(self):
		"""
		Opens a popup to load the RunQueue-items from a file.
		Loading itself is done by the run_queue itself.
		"""
		if self._run_queue.get_running_configuration_count() > 0:
			msg_box = QtWidgets.QMessageBox()
			msg_box.setIcon(QtWidgets.QMessageBox.Icon.Warning)
			msg_box.setWindowTitle("Cannot load queue from file")
			msg_box.setText("Cannot load queue from file while there are still running items in the queue.")
			msg_box.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
			msg_box.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Ok)
			msg_box.setWindowFlags(QtCore.Qt.WindowType.WindowStaysOnTopHint)
			msg_box.show()
			msg_box.activateWindow()
			msg_box.exec()
			return


		file_path, _ = QtWidgets.QFileDialog.getOpenFileName(
			None, "Load config", QtCore.QDir.currentPath(), "RunQueue files (*.rq)") #type:ignore

		file_path = file_path.replace("/", os.sep) #Get the actual path (qt uses / as separator)

		self.load_from_file(file_path)

	def get_n_processes(self) -> int:
		"""Returns the number of processes to use for the queue.
		"""
		return self._run_queue.get_n_processes()

	def set_n_processes(self, n_processes : int) -> None:
		"""Sets the number of processes to use for the queue.

		Args:
			n_processes (int): The number of processes to use for the queue
		"""
		self._run_queue.set_n_processes(n_processes)

	def save_to_file_popup(self):
		"""Opens a popup to save the queue to a file.

		Save-handling is done by the run_queue itself.

		NOTE: if we make more models based on the run_queue, we should probably move this to a shared base-class
		"""
		try:
			try:
				save_dict = self._run_queue.get_queue_contents_dict(save_running_as_stopped=False)
			except RunQueueHasRunningItemsException as exception:
				log.warning("")
				#Ask if user wants to save anyway, in the savefile, all running items will be marked as "stopped"
				#and the queue will be saved
				msg_box = QtWidgets.QMessageBox()
				msg_box.setIcon(QtWidgets.QMessageBox.Icon.Warning)
				msg_box.setWindowTitle("Save queue to file anyway?")
				msg = f"Problem when saving queue to file: {type(exception).__name__}: {exception}"
				log.warning(msg)
				msg_box.setText("<b>Do you want to set save mode to allow saving running configurations?</b>")
				msg_box.setInformativeText(
					f"<b>{type(exception).__name__}:</b> {exception} <br><br>"
					"We can enable saving running configurations, but this will cause the running configurations to "
					"be saved with a 'stopped'-state.  "
					"Alternatively, you can cancel all running items or wait for them to finish and try again. "

				)
				msg_box.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
				msg_box.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
				msg_box.setWindowFlags(QtCore.Qt.WindowType.WindowStaysOnTopHint)
				msg_box.show()
				msg_box.activateWindow()
				msg_box.exec()

				if msg_box.result() == QtWidgets.QMessageBox.StandardButton.Yes: #Load and allow running items
					save_dict = self._run_queue.get_queue_contents_dict(save_running_as_stopped=True)
				else:
					log.info("User cancelled saving queue to file.")
					return

			file_path, _ = QtWidgets.QFileDialog.getSaveFileName(
				None, "Save config", QtCore.QDir.currentPath(), "RunQueue files (*.rq)") #type:ignore

			if file_path is None or file_path == "":
				log.info("User cancelled saving queue to file.")
				return

			with open(file_path, "wb") as save_file:
				dill.dump(save_dict, save_file)

			log.info(f"Saved RunQueue to file {file_path}.")

		except Exception as exception: #pylint: disable=broad-exception-caught
			msg = f"{type(exception).__name__}: {exception}"
			log.warning(f"Could not save RunQueue to file - {msg}")
			msg_box = QtWidgets.QMessageBox()
			msg_box.setIcon(QtWidgets.QMessageBox.Icon.Warning)
			msg_box.setWindowTitle("Could not save RunQueue to file")
			msg_box.setInformativeText(msg)
			msg_box.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
			msg_box.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Ok)
			msg_box.setWindowFlags(QtCore.Qt.WindowType.WindowStaysOnTopHint)
			msg_box.show()
			msg_box.exec()
			return


		#Check if user cancelled




	def rowCount(self,
	      	parent: typing.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex, None] = None #pylint: disable=unused-argument
		) -> int:
		return len(self._cur_item_dict_id_order) #NOTE: the _cur_item_dict_id order determines what items are shown
			# even when more items are in the _cur_item_dict


		# return super().rowCount(parent)
	def columnCount(self,
			parent: typing.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex, None] = None #pylint: disable=unused-argument
		) -> int:
		return len(self.column_names)


	def _handle_run_queue_insertion(self, new_item_ids : list[int], new_items_dict : typing.Dict[int, RunQueueItem]):
		"""
		Handles the insertion of new item(s) in the runQueue. Should be linked to the allItemsDictInsertion signal
		of the attached runQueueManager. On item insertion, updates the table-model.

		Args:
			new_item_ids (list[int]): list of ids of the new items
			new_items_dict (typing.Dict[int, RunQueueItem]): a copy of the all_items_dict with the new item added
		"""
		if (len(new_item_ids) - len(new_items_dict)) != len(self._cur_run_queue_item_dict_copy):
			log.warning("New item(s) added to runQueue, but the updated length of the runQueue is different from the currently"
	     		"known length of the table-model plus the amount of added items. The runqueue seems to be out of sync...")

		self._cur_run_queue_item_dict_copy = new_items_dict #Update the copy of the runQueue

		#Also look for the new item-order (sort the dict keys = sort by id)
		new_item_dict_id_order = list(new_items_dict.keys())
		new_item_dict_id_order.sort()

		insertion_points = [new_item_dict_id_order.index(id) for id in new_item_ids] #Check where the new items were inserted
		#Make the insertion points negative so we can iteratively add items while the insertion points will be determined

		cur_insertion_point_index = 0
		while True: #Group all consecutive insertions together ([1,2,3,5] -> [1,2,3] and [5])
			start_insertion_point_idx = cur_insertion_point_index
			stop_insertion_point_idx  = cur_insertion_point_index
			for i, insertion_point in enumerate(insertion_points[cur_insertion_point_index+1:]): #All next insertion points
				last_insertion_point = insertion_points[cur_insertion_point_index]
				if (last_insertion_point + i + 1) == insertion_point:
					stop_insertion_point_idx  = insertion_point
					cur_insertion_point_index += 1
				else:
					break

			self.beginInsertRows(QtCore.QModelIndex(), start_insertion_point_idx, stop_insertion_point_idx)
			for i in range(start_insertion_point_idx, stop_insertion_point_idx +1): #Insert the ids into the id-order-list
				insert_loc  = insertion_points[i]
				self._cur_item_dict_id_order.insert(insert_loc, new_item_dict_id_order[insert_loc])
			self.endInsertRows()
			cur_insertion_point_index += 1
			if cur_insertion_point_index >= len(insertion_points): #If inserted all items
				break

		if self._cur_item_dict_id_order != new_item_dict_id_order:
			log.warning("Item order in model does not match item order in runQueue")

	def force_stop_all_running(self):
		"""Force stop all items in the runqueue, also stops autoprocessing

		NOTE: all unsaved progress will be lost(!!!)
		"""
		self._run_queue.force_stop_all_running()

	def _handle_run_queue_deletion(self, deleted_item_ids : list[int], new_items_dict : typing.Dict[int, RunQueueItem]):
		"""
		Handles the deletetion of item(s) in the runQueue

		Args:
			deleted_item_ids (list[int]): list of ids of the deleted items
			new_items_dict (typing.Dict[int, RunQueueItem]): a copy of the all_items_dict with the new item added
		"""
		if (len(new_items_dict) - len(deleted_item_ids)) != len(self._cur_run_queue_item_dict_copy):
			log.warning("Item(s) removed from runQueue, but the updated length of the runQueue is different from the currently"
	     		"known length of the table-model minus the amount of removed items. The runqueue seems to be out of sync...")

		# self._cur_run_queue_item_dict_copy = new_items_dict #Update the copy of the runQueue

		#Also look for the new item-order (sort the dict keys = sort by id)
		new_item_dict_id_order = list(new_items_dict.keys())
		new_item_dict_id_order.sort() #

		deletion_points = [self._cur_item_dict_id_order.index(id) for id in deleted_item_ids] #Check where the new items
			#were deleted
		deletion_points = deletion_points[::-1] #So we can delete items from the back to the front so indexes don't change
		#Make the insertion points negative so we can iteratively add items while the insertion points will be determined

		cur_deletion_point_index = 0
		while True : #Group all consecutive deletions together ([5,3,2,1] -> [5] [3,2,1])
			start_deletion_point_idx = cur_deletion_point_index
			stop_deletion_point_idx  = cur_deletion_point_index
			for i, insertion_point in enumerate(deletion_points[cur_deletion_point_index+1:]): #All next insertion points
				last_insertion_point = deletion_points[cur_deletion_point_index]
				if (last_insertion_point + i - 1) == insertion_point:
					stop_deletion_point_idx  = insertion_point
					cur_deletion_point_index += 1
				else:
					break

			# self.beginInsertRows(QtCore.QModelIndex(), start_deletion_point_idx, stop_deletion_point_idx)
			self.beginRemoveRows(QtCore.QModelIndex(), start_deletion_point_idx, stop_deletion_point_idx)
			for i in range(start_deletion_point_idx, stop_deletion_point_idx +1): #Insert the ids into the id-order-list
				delete_loc  = deletion_points[i]
				delete_id = self._cur_item_dict_id_order[delete_loc]

				#Remove item from dict and id-order-list
				del self._cur_run_queue_item_dict_copy[delete_id]
				del self._cur_item_dict_id_order[delete_loc]

				# self._cur_item_dict_id_order.insert(delete_loc, new_item_dict_id_order[delete_loc])
			self.endRemoveRows()
			cur_deletion_point_index += 1
			if cur_deletion_point_index >= len(deletion_points): #If inserted all items
				break

		if self._cur_item_dict_id_order != new_item_dict_id_order:
			log.warning("Item order in model after deletion does not match item order in runQueue")

	def _queue_changed(self, queue_copy):
		self._cur_queue_copy = queue_copy
		# self.layoutChanged.emit()
		log.debug(f"The queue order changed to {', '.join([str(i) for i in queue_copy])}. "
	    	"Now emitting datachanged for column...")

		row = self.column_positions["status"]
		self.dataChanged.emit(
			self.index(row, 0),
			self.index(row, self.columnCount()),
			[QtCore.Qt.ItemDataRole.DisplayRole, QtCore.Qt.ItemDataRole.EditRole] #Change the displayed queue-position
		) #TODO: Only emit changed. rows


	def _run_item_changed(self, changed_item_id : int, new_item_copy : RunQueueItem):
		# row = list(self._cur_run_queue_item_dict_copy.keys()).index(changed_item_id)
		row = self.get_index_row_by_id(changed_item_id)

		assert(changed_item_id in self._cur_run_queue_item_dict_copy), "ID of update-item not in currently known itemList"
		self._cur_run_queue_item_dict_copy[changed_item_id] = new_item_copy
		self.dataChanged.emit(
			self.index(row, 0),
			self.index(row, self.columnCount())
		) #TODO: Only emit for the changed rows

	def get_index_row_by_id(self, item_id : int) -> int:
		"""Get the index-row of the item with the given id

		Args:
			item_id (int): The id of the item to get the index for

		Returns:
			row (int): The row of the item with the given id
		"""
		if item_id in self._cur_item_dict_id_order:
			# id_list = list(self._cur_run_queue_item_dict_copy.keys())
			# id_list.sort()
			# return id_list.index(item_id)
			return self._cur_item_dict_id_order.index(item_id)
		else:
			return -1

	def get_index_by_id(self, item_id : int, column : int):
		"""Get the row if the index by id, and construct the index from that row and the given column"""
		return self.index(self.get_index_row_by_id(item_id), column)



	def set_highlight_by_index(self, index: QtCore.QModelIndex) -> None:
		"""Set the highlighted item by its index in the model
		Args:
			index (QtCore.QModelIndex): The index of the item to highlight
		"""
		if index.isValid():
			new_id = index.data(role=RunQueueTableModel.CustomDataRoles.IDRole)
			self._prev_highlighted_id = self._highlighted_id
			self._highlighted_id = new_id

			self.dataChanged.emit(
				self.index(index.row(), 0),
				self.index(index.row(),self.columnCount()), #update column
				[QtCore.Qt.ItemDataRole.FontRole] #Only update font as it is the only thing that changes due to highlighting
			)

			self.itemHighlightIdChanged.emit(new_id) #Emit signal to UI
		else:
			log.info("Invalid index given to set_highlight_by_index, resetting highlight")
			self.itemHighlightIdChanged.emit(None) #Emit signal to UI
			self._highlighted_id = -1

		if self._prev_highlighted_id is not None: #Also clear boldness of the previous item
			self.dataChanged.emit(
				self.index(self.get_index_row_by_id(self._prev_highlighted_id), 0),
				self.index(self.get_index_row_by_id(self._prev_highlighted_id), self.columnCount()),
				[QtCore.Qt.ItemDataRole.FontRole] #Only update font as it is the only thing that changes due to highlighting
			)

	def set_highligh_by_id(self, highlight_item_id : int | None) -> None:
		"""Set the highlighted item by its item id
		Args:
			id (int): The id of the item to highlight or None/--1 to clear the highlight
		"""
		self._prev_highlighted_id = self._highlighted_id
		if highlight_item_id is None or highlight_item_id < 0:
			self._highlighted_id = -1
		else:
			self._highlighted_id = highlight_item_id
			self.dataChanged.emit(
				self.get_index_by_id(self._highlighted_id, 0),
				self.get_index_by_id(self._highlighted_id, self.columnCount()), #update column
				[QtCore.Qt.ItemDataRole.FontRole] #Only update font as it is the only thing that changes due to highlighting
			) #TODO: Only update the row with the new/old id
		if self._prev_highlighted_id is not None: #Also clear boldness of the previous item
			self.dataChanged.emit(
				self.get_index_by_id(self._prev_highlighted_id, 0),
				self.get_index_by_id(self._prev_highlighted_id, self.columnCount()), #update column
				[QtCore.Qt.ItemDataRole.FontRole] #Only update font as it is the only thing that changes due to highlighting
			)
		self.itemHighlightIdChanged.emit(highlight_item_id) #Emit signal to UI

	def hightlighted_id(self) -> int | None:
		"""return the currently highlighted item id in the model"""
		return self._highlighted_id


	#TODO: use RunqueueItemStatus instead of id to get options for ID -> otherwise we have to "ask"remote
	#server for item status every time we select a row -> might not be desireable
	def get_actions(self, index : QtCore.QModelIndex) -> typing.List[RunQueueItemActions]:
		"""Retrieve the possible actions for a given index (queue item), such as delete, move up in queue, etc.

		Args:
			index (QtCore.QModelIndex): The index of the item for which to retrieve the possible actions

		Returns:
			typing.List[QueueItemActions]: A list of possible actions for the given index.
		"""
		if index.isValid():
			status = self.get_item_status(index)
			if status is not None:
				return RunQueue.get_actions_from_status(status)
			else:
				return []
		else:
			return []

	def add_to_queue(self, name, config) -> None:
		"""Addd a new item to the queue with the given name and config"""
		self._run_queue.add_to_queue(name, config)

	def do_action_by_id(self, item_id : int, action : RunQueueItemActions) -> None:
		"""Perform the given action on the item with the given id

		Args:
			item_id (int): The id of the item to perform the action on
			action (RunQueueItemActions): The action to perform
		"""
		self._run_queue.do_action_for_id(item_id, action)

	def do_action(self, index : QtCore.QModelIndex, action : RunQueueItemActions) -> None:
		"""Determine the id of the item at the given index and perform the given action on it

		Args:
			index (QtCore.QModelIndex): The index of the item to perform the action on
			action (RunQueueItemActions): The action to perform
		"""
		target_id = index.data(role=RunQueueTableModel.CustomDataRoles.IDRole)
		log.debug(f"Item id is: {target_id}")
		if index.isValid():
			log.debug(f"Performing action {action} on item with id {target_id}")
			self._run_queue.do_action_for_id(target_id, action)


	def get_item_status(self, index : QtCore.QModelIndex) -> RunQueueItemStatus | None:
		"""Get the item status by index

		Args:
			index (QtCore.QModelIndex): the index for which to fetch the status

		Returns:
			RunQueueItemStatus | None: The status of the item, or None if the index is invalid
		"""
		if index.isValid():
			target_id = self._cur_item_dict_id_order[index.row()]
			return self._cur_run_queue_item_dict_copy[target_id].status
		else:
			return None

	def setData(self, index: QtCore.QModelIndex, value: typing.Any, role: int | None = None) -> bool:
		"""The setData function, in this case only implements the action role. Makes it possible to perform an action
		(e.g. delete, move up, etc.) on an item in the queue with proxy-models in between.

		Calls to setData are propagated through all proxy-models to the underlying model (RunQueueTableModel).

		NOTE: this method differs somewhat from the normal implementation of setData in that it doesn't directly
		emit a dataChanged signal if the set was succesful. Instead, it attempts to do the action on the runqueue
		and if that succeeds, the runQueue will emit a signal that will cause the model to update itself.
		"""
		log.debug(f"setData called for index {index} with value {value} and role {role}")
		try:
			if index.isValid():
				# if role == QtCore.Qt.ItemDataRole.EditRole:
				# 	if index.column() == 0:
				# 		self._cur_run_list_copy[index.row()].name = value
				# 		self.dataChanged.emit(index, index)
				# 		return True
				log.debug("Index valid, not checking type and action")
				if role == RunQueueTableModel.CustomDataRoles.ActionRole:
					if not isinstance(value, RunQueueItemActions):
						log.error(f"Invalid role {role} for setData for index {index} in RunQueueTableModel")
						return False
					log.debug("Now actually doing action")
					self.do_action(index, value)
					log.debug("Done doing action")
					#Emit whole row:
					return False
				else:
					log.error(f"Invalid role {role} for setData for index {index} in RunQueueTableModel")
		except Exception as exception: # pylint: disable=broad-except
			log.error(f"Failed to set data for index {index}: {exception}")
		return False

	def data(self, index: QtCore.QModelIndex, role: int | None = None) -> typing.Any:
		try:
			item_id = self._cur_item_dict_id_order[index.row()]
			item = self._cur_run_queue_item_dict_copy[item_id]

			if role == QtCore.Qt.ItemDataRole.DisplayRole:
				key :str = self.column_names[index.column()][0]
				key = key.lower()
				attr = getattr(item, key)

				if key == "config":
					return "-"
				if key == "status": #Display position in queue
					cur_id = list(self._cur_run_queue_item_dict_copy.keys())[index.row()]
					if item.status == RunQueueItemStatus.Queued and cur_id in self._cur_queue_copy:
						return f"In Queue: {self._cur_queue_copy.index(cur_id)+1}/{len(self._cur_queue_copy)}"
					elif item.status == RunQueueItemStatus.Queued:
						#If the item says it is queued, but it is not in the queue, the model is out of sync
						# with the current state of the queue and or the runqueue-items
						return "ERROR: MODEL OUT OF SYNC"
					else: #Convert enum to string
						return str(item.status.name)
				if (key == "dt_added") or (key == "dt_done") or (key == "dt_started"):
					if attr:
						# print(key, attr)
						return attr.strftime("%Y-%m-%d %H:%M:%S")
				return attr


			elif role == QtCore.Qt.ItemDataRole.DecorationRole:
				if index.column() == 0:
					return self._icon_dict.get(item.status, None)
				else:
					return None
			elif role == QtCore.Qt.ItemDataRole.ToolTipRole:
				ret = str(self.data(index, QtCore.Qt.ItemDataRole.DisplayRole))
				#Go over ret lines and if they are too long, split them
				ret_lines = ret.split("\n")
				for i, line in enumerate(ret_lines):
					if len(line) > 80: #If line is too long, split it
						ret_lines[i] = "\n".join(textwrap.wrap(line, 80))
				return "\n".join(ret_lines)
			elif role == QtCore.Qt.ItemDataRole.FontRole: #Font role: highlight if item is highlighted
				if item.item_id == self._highlighted_id:
					return self._highlight_font
			elif role == RunQueueTableModel.CustomDataRoles.IDRole:
				return item.item_id
			elif role == RunQueueTableModel.CustomDataRoles.ActionRole:
				return self.get_actions(index)
			elif role == RunQueueTableModel.CustomDataRoles.StatusRole:
				return (item.status,) #Returned as tuple, because otherwise the return value is converted to an int (?)
				# return "kaas"
			return None

		except Exception as exception: # pylint: disable=broad-except
			log.error(f"Failed to get data for index {index}: {exception}")

	def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int | None = None) -> typing.Any:
		if role == QtCore.Qt.ItemDataRole.DisplayRole:
			if orientation == QtCore.Qt.Orientation.Horizontal and section > 0:
				return self.column_names[section][1]
		return None



"""
Implements the RunQueueWidget class which contains a treeview and several buttons to control the underlying queue-model
as well as (indirectly) the underlying RunQueue
"""

import logging

from PySide6 import QtCore, QtGui, QtWidgets
from pyside6_utils.utility.catch_show_exception_in_popup_decorator import \
    catch_show_exception_in_popup_decorator

from configurun.classes.run_queue import RunQueueItemActions
from configurun.app.models.run_queue_table_model import RunQueueTableModel
from configurun.app.ui.run_queue_widget_ui import Ui_RunQueueWidget
from configurun.classes.run_queue import RunQueueItemStatus

log = logging.getLogger(__name__)



class RunQueueWidget(QtWidgets.QWidget):
	"""
	A wrapper-widget for a RunQueueTreeView with some buttons to control the queue (start, stop, move up, move down,
	delete,	cancel etc.)
	"""
	def __init__(self, widget : QtWidgets.QWidget) -> None:
		super().__init__()
		self.ui = Ui_RunQueueWidget() #pylint: disable=invalid-name
		self.ui.setupUi(widget)
		self._widget = widget

		self.runQueueTreeView = self.ui.runQueueTreeView #pylint: disable=invalid-name #Make it easier to access treeview

		#=============Link buttons to functions================
		self.ui.MoveUpInQueueBtn.clicked.connect(
			lambda *_: self.runQueueTreeView.do_action_on_selection(RunQueueItemActions.MOVEUP))
		self.ui.MoveDownInQueueBtn.clicked.connect(
			lambda *_: self.runQueueTreeView.do_action_on_selection(RunQueueItemActions.MOVEDOWN))
		self.ui.CancelStopButton.clicked.connect(
			self.cancel_stop_button_pressed)
		self.ui.DeleteButton.clicked.connect(
			lambda *_: self.runQueueTreeView.do_action_on_selection(RunQueueItemActions.DELETE))
		self.ui.StartRunningQueueBtn.clicked.connect(
			self.toggle_queue_autoprocessing)

		self._action_btn_dict = {
			RunQueueItemActions.MOVEUP: self.ui.MoveUpInQueueBtn,
			RunQueueItemActions.MOVEDOWN: self.ui.MoveDownInQueueBtn,
			RunQueueItemActions.CANCEL: self.ui.CancelStopButton, #TODO: is it a good idea to have cancel and stop be the
			 # same button? Maybe make difference more apparent to user
			 # (cancel = dequeue, stop = stop if currently running and dequeue)
			RunQueueItemActions.STOP: self.ui.CancelStopButton,
			RunQueueItemActions.DELETE: self.ui.DeleteButton
		}

		self.queue_model_connections = [] #Connections to the queue model for updating the UI
		self._run_queue_table_model : RunQueueTableModel | None = None #The used run-queue model

		self.ui.toolButton.clicked.connect(self.settings_clicked)


	def settings_clicked(self):
		"""Called when a open-settings-button is clicked. Opens the settings menu."""
		cur_mouse_pos = QtGui.QCursor.pos()
		self.popup_queue_settings_context_menu(cur_mouse_pos)

	def get_current_view_filter_menu(self) -> QtWidgets.QMenu:
		"""Returns the current view filter-menu for the queue view. E.g. a checkmark for all statuses that
		are being shown, and none for all statuses that are being filtered out from the tree-view

		Returns:
			QtWidgets.QMenu: The current view filter-settings for the queue view
		"""
		menu = QtWidgets.QMenu()
		cur_filtered = self.runQueueTreeView.get_item_status_filter() #Get all filtered statuses
		for status in RunQueueItemStatus:
			action = QtGui.QAction(status.name, menu)
			action.setCheckable(True)
			action.setChecked(status not in cur_filtered)
			action.triggered.connect(
				lambda *_, status=status: self.runQueueTreeView.set_whether_status_filtered(
					status, not(status in cur_filtered)) #pylint: disable=superfluous-parens
				)
			menu.addAction(action)

		return menu


	def save_to_file_popup(self):
		"""Shows a popup to save the current queue to a file."""
		cur_model = self._run_queue_table_model
		assert isinstance(cur_model, RunQueueTableModel), f"Can't save Queue for non-RunQueueModel of type {type(cur_model)}"
		cur_model.save_to_file_popup()

	def load_from_file_popup(self):
		"""Shows a popup to load a queue from a file."""
		cur_model = self._run_queue_table_model
		assert isinstance(cur_model, RunQueueTableModel), f"Can't load Queue for non-RunQueueModel of type {type(cur_model)}"
		cur_model.load_from_file_popup()


	def get_queue_settings_context_menu(self) -> QtWidgets.QMenu:
		"""Returns the context menu for the current RunQueue, e.g. which items are filtered in the tree view
		and whether the queue is backed up to a file.

		Returns:
			QtWidgets.QMenu: The context menu for the current RunQueue
		"""
		menu = QtWidgets.QMenu()
		# filter_menu = menu.addMenu("View Filter")
		filter_menu = self.get_current_view_filter_menu()
		filter_menu.setTitle("View Filter")
		filter_menu.setIcon(QtGui.QIcon(":/Icons/icons/actions/filter.png"))
		menu.addMenu(filter_menu)


		backup_action = menu.addAction("Backup...")
		backup_action.setIcon(QtGui.QIcon(":/Icons/icons/actions/document-save-as.png"))
		cur_model = self._run_queue_table_model
		if cur_model is not None:
			backup_action.triggered.connect(self.save_to_file_popup)
		else:
			backup_action.setEnabled(False)


		load_action = menu.addAction("Load...")
		load_action.setIcon(QtGui.QIcon(":/Icons/icons/actions/document-open.png"))
		if cur_model is not None:
			load_action.triggered.connect(self.load_from_file_popup)
		else:
			load_action.setEnabled(False)

		n_processes_action = menu.addAction("Set number of processes...")
		n_processes_action.setIcon(QtGui.QIcon(":/Icons/icons/status/network-receive.png"))
		n_processes_action.triggered.connect(self.set_n_processes_popup)

		return menu

	def set_n_processes_popup(self):
		"""Shows a popup to set the number of processes to use in the run queue."""
		#Ask user for integer
		#Set number of processes to that integer

		if self._run_queue_table_model is None:
			QtWidgets.QMessageBox.warning(
				self._widget,
				"Can't set number of processes",
				"Can't set number of processes, as RunQueueModel is not set"
			)
			return
		cur_n_processes = self._run_queue_table_model.get_n_processes()


		integer, pressed_ok = QtWidgets.QInputDialog.getInt(
			self._widget,
			"Set number of processes",
			f"Set Number of processes: <br>current={cur_n_processes} unlimited=-1",
			value=cur_n_processes,
			minValue=-1,
			maxValue=30,
			step=1
		)
		if pressed_ok:
			self._run_queue_table_model.set_n_processes(integer)

	def popup_queue_settings_context_menu(self, global_position : QtCore.QPoint) -> None:
		"""Creates a menu with settings for the queue and shows it at the given position.
		e.g. filter by status.

		Args:
			global_position (QtCore.QPoint): The position to show the menu at.
		"""
		menu = self.get_queue_settings_context_menu()
		menu.exec(global_position)



	def cancel_stop_button_pressed(self):
		"""Either cancels or stops the currently selected item in the queue, depending on its status
		"""
		# cur_model = self.queue_view.model()
		# assert isinstance(cur_model, RunQueueTableModel), "Can't get options for non-MLQueueModel"
		selection = self.runQueueTreeView.currentIndex()

		actions = selection.data(RunQueueTableModel.CustomDataRoles.ActionRole)
		if RunQueueItemActions.STOP in actions:
			# self.queue_view.do_action_on_index(RunQueueItemActions.STOP, selection)
			self.runQueueTreeView.confirm_stop_index(selection)
		else:
			self.runQueueTreeView.do_action_on_index(RunQueueItemActions.CANCEL, selection)



	def _autoqueue_btn_set_state(self, is_running : bool):
		"""Set the state of the autoqueue button (start/stop autoqueueing)"""
		self.ui.StartRunningQueueBtn.setChecked(is_running)
		self.ui.StartRunningQueueBtn.setToolTip("Stop autorunning" if is_running else "Start autorunning")

	@catch_show_exception_in_popup_decorator
	def toggle_queue_autoprocessing(self):
		"""Toggles the queue autoprocessing."""
		cur_model = self._run_queue_table_model
		assert isinstance(cur_model, RunQueueTableModel)


		autoprocessing_enabled = cur_model.is_autoprocessing_enabled()
		self._autoqueue_btn_set_state(autoprocessing_enabled)

		if autoprocessing_enabled:
			if cur_model.get_running_configuration_count() <= 0:
				cur_model.stop_autoprocessing()
				self._autoqueue_btn_set_state(False) #Set button state to false TODO: intermediary state when stopping?
				return
			else:
				confirm_dialog = QtWidgets.QMessageBox()
				#Creat a dialog with "Wait for processes to finish" and "Force Stop" and "Cancel" buttons
				confirm_dialog = QtWidgets.QMessageBox()
				confirm_dialog.setText("You are about to stop the autoprocessing process, do you want to cancel all "
			   		"currently running processes as well?")
				#Create a dialog with "Wait for processes to finish" and "Force Stop" and "Cancel" buttons
				#If "Wait and Stop" is pressed, stop autoqueueing
				#If "Force Stop" is pressed, stop all items currently running in the queue and stop autoqueueing
				#If "Cancel" is pressed, do nothing
				stop_autoqueue_btn = \
					confirm_dialog.addButton("Stop Autoqueueing", QtWidgets.QMessageBox.ButtonRole.AcceptRole) #0
				force_stop_btn = \
					confirm_dialog.addButton("Force Stop All", QtWidgets.QMessageBox.ButtonRole.NoRole) #1
				cancel_btn = \
					confirm_dialog.addButton("Cancel", QtWidgets.QMessageBox.ButtonRole.RejectRole) #2

				#Set default button to "Wait and Stop"
				confirm_dialog.setDefaultButton(stop_autoqueue_btn)

				confirm_dialog.setWindowTitle("Stop queue?")
				confirm_dialog.setWindowIcon(QtGui.QIcon("resources/icons/icon.ico"))
				confirm_dialog.setWindowFlags(QtCore.Qt.WindowType.WindowStaysOnTopHint)
				confirm_dialog.show()
				confirm_dialog.activateWindow()
				confirm_dialog.exec()


				if confirm_dialog.clickedButton() == cancel_btn: #If cancel is pressed (third button = index 2)
					return
				elif confirm_dialog.clickedButton() == stop_autoqueue_btn: #If wait stop
					cur_model.stop_autoprocessing()
					self._autoqueue_btn_set_state(False) #Set button state to false
					return
				elif confirm_dialog.clickedButton() == force_stop_btn:
					#1 last check if user really wants to force stop (use one-liner messagebox here)
					confirm_window = QtWidgets.QMessageBox()
					confirm_window.setText("<b>Are you sure you want to force stop all currently running processes?</b>")
					confirm_window.setInformativeText("This will forcefully stop all currently running processes. "
						"All unsaved progress will be lost. This action cannot be undone.")
					confirm_window.setWindowTitle("Force stop?")

					confirm_window.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes
						| QtWidgets.QMessageBox.StandardButton.No)
					confirm_window.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
					if confirm_window.exec() == QtWidgets.QMessageBox.StandardButton.No:
						return
					cur_model.force_stop_all_running()
		else:
			cur_model.start_autoprocessing()

	def set_model(self, model : RunQueueTableModel):
		"""Resets all signals and connections and invalidates the current data in the view.
		"""
		assert isinstance(model, RunQueueTableModel), "Can't set non-MLQueueModel as model"
		if len(self.queue_model_connections) > 0:
			for signal_connection in self.queue_model_connections:
				signal_connection.disconnect()
			self.queue_model_connections = []

		# cur_model = self.queue_view.model()
		self.runQueueTreeView.setModel(model)
		self._run_queue_table_model = model

		if not isinstance(model, RunQueueTableModel):
			self.update_available_options()
			log.warning("MLQueueWidget: Model is not of type MLQueueModel. Cannot setup new model.")
			return

		self.queue_model_connections.append( #When any data changes, update the available options
			model.dataChanged.connect(lambda *_: self.update_available_options())
		)

		#=============== Subscribe to signals =================
		self.queue_model_connections.append(
			model.layoutChanged.connect(lambda *_: self.update_available_options())
		) #When any rows are removed, update the available options (maybe selected item changed)
		self.queue_model_connections.append(
			self.runQueueTreeView.selectionModel().selectionChanged.connect(lambda *_: self.update_available_options())
		) #When the selection changes, update the available options
		self.queue_model_connections.append(
			self.runQueueTreeView.selectionModel().currentChanged.connect(lambda *_: self.update_available_options())
		) #When the selection changes, update the available options
		self.queue_model_connections.append(
			model.autoProcessingStateChanged.connect(self._autoqueue_btn_set_state)
		)

		self._autoqueue_btn_set_state(model.is_autoprocessing_enabled())
		self.update_available_options()


	@catch_show_exception_in_popup_decorator
	def force_stop_queue(self):
		""" Stops the auto-requeueing of items in the RunQueue """
		raise NotImplementedError("MLQueueWidget: stop_queue not implemented yet")

	def update_available_options(self):
		""" Updates the available options in the user interface. """
		index = self.runQueueTreeView.currentIndex()
		# log.debug(f"Updating available options for selection {index}")
		if not index.isValid():
			cur_available_actions = []
		else:
			cur_available_actions = index.data(RunQueueTableModel.CustomDataRoles.ActionRole) #Retrieve the available actions


		for btn in self._action_btn_dict.values():
			btn.setEnabled(False)

		if cur_available_actions is not None:
			for action in cur_available_actions:
				if action in self._action_btn_dict:
					self._action_btn_dict[action].setEnabled(True)




def run_example_app():
	""" Run an example of the run_queue widget with some nonsense data and the example run function"""
	#pylint: disable=import-outside-toplevel
	from configurun.examples.example_target_function import example_target_function
	from configurun.classes.run_queue import RunQueue

	app = QtWidgets.QApplication([])
	window = QtWidgets.QMainWindow()
	window.resize(800, 600)
	central_widget = QtWidgets.QWidget()

	queue_widget = RunQueueWidget(central_widget)
	run_queue = RunQueue(
		target_function=example_target_function
	)
	queue_model = RunQueueTableModel(run_queue=run_queue)

	queue_widget.set_model(queue_model)

	for i in range(10):
		run_queue.add_to_queue(
			f"Testitem {i}",
			config={"test": i}, #nonsense config #type:ignore
		)

	window.setCentralWidget(central_widget)

	window.show()
	app.exec()


if __name__ == "__main__":
	logging.getLogger('matplotlib').setLevel(logging.INFO)
	logging.getLogger('PySide6').setLevel(logging.DEBUG)
	formatter = logging.Formatter("[{pathname:>90s}:{lineno:<4}] {levelname:<7s}   {message}", style='{')
	handler = logging.StreamHandler()
	handler.setFormatter(formatter)
	logging.basicConfig(
		handlers=[handler],
		level=logging.DEBUG
	) #Without time
	root = logging.getLogger()
	root.handlers = [handler]
	run_example_app()



"""
Contains the main window class, which is a window which provides the user with several tools to
edit/manage/run machine learning settings.

Also contains OptionsSource, which is used to determine if the current file should be saved to a file or to the queue.
"""
import logging
import os
import typing
from enum import Enum

import dill
from PySide6 import QtCore, QtGui, QtWidgets
from pyside6_utils.models import FileExplorerModel
from pyside6_utils.utility.catch_show_exception_in_popup_decorator import \
    catch_show_exception_in_popup_decorator
from pyside6_utils.widgets import DataClassTreeView
from pyside6_utils.widgets.delegates import DataclassEditorsDelegate
from pyside6_utils.widgets.frameless_mdi_window import FramelessMdiWindow

from configurun.classes.run_queue import \
	ConfigurationIsFirmException, RunQueue, WORKSPACE_RUN_QUEUE_SAVE_NAME, WorkspaceInUseException
from configurun.configuration.configuration import Configuration
from configurun.configuration.configuration_model import (
    ConfigurationModel, NoClassTypesError, OptionTypesMismatch,
    UnkownOptionClassError)
from configurun.app.models.run_queue_console_model import \
    RunQueueConsoleModel
from configurun.app.models.run_queue_table_model import RunQueueTableModel
from configurun.app.ui.main_window_ui import Ui_MainWindow


log = logging.getLogger(__name__)

APP_NAME = "Configurun" #The name of the app, used for the settings file

class OptionsSource(Enum):
	"""
	Source of the options. Used to determine whether the options should be saved to a file or to the queue.
	"""
	FILE = 0
	QUEUE = 1


class MainWindow(QtCore.QObject):
	"""
	The main QT window for this app which provides the user with several tools to edit/manage/run machine learning
	settings.

	Should be provided with:
		- A configuration model - manages the creation of new configurations & the ui
		- A run queue - manages the running of the configurations
		- A window - the main window in which the app should be built
		- workspace_path (str, optional) - the default path to use for the configuration, logfiles etc. If empty, or
			folder does not exist, defaults to ~/Configurun/configurations/
		- settings_in_workspace_path (bool, optional) - Whether to store the settings in the workspace path or in the
			default QSettings location. Defaults to True
	"""
	def __init__(self,
	      		configuration_model : ConfigurationModel,
				run_queue : RunQueue,
				window : QtWidgets.QMainWindow,
				workspace_path : str = "",
				settings_in_workspace_path : bool = True
			) -> None:
		"""
		Args:
			configuration_model (ConfigurationModel): The configuration model which manages updating the ui and creating
			run_queue (RunQueue): The runqueue which manages running the configurations
			window (QtWidgets.QMainWindow): The window in which the app should be built
			workspace_path (str, optional): The base output-path used for the configurations, logfiles etc.
				If empty, or folder does not exist, defaults to ~/Configurun/configurations/
			settings_in_workspace_path (bool, optional): Whether to store the settings in the workspace path or in the default
				QSettings location. Defaults to True
		"""
		super().__init__()
		self.ui = Ui_MainWindow() # pylint: disable=C0103
		self.ui.setupUi(window)

		self.window = window
		self._cur_source = None
		self._queue_source_id = -1 #The id of the queue item which is currently selected (if cur_source == QUEUE)
		self._default_splitter_states = {
			splitter.objectName() : splitter.saveState() \
				for splitter in self.window.findChildren(QtWidgets.QSplitter)
		} #Save all splitter states (to be able to reset them later

		self.ui.saveToQueueItemBtn.setHidden(True) #Hide the save to queue button until queue-item is selected
		#====================== Base variables ===================
		self.set_run_queue(run_queue)
		self.ui.runQueueWidget.set_model(self.run_queue_table_model)
		self.ui.runQueueWidget.runQueueTreeView.set_double_click_callback(self.run_queue_widget_item_double_click)


		self._workspace_path = workspace_path
		if workspace_path is None or len(workspace_path) == 0 or not os.path.isdir(workspace_path):
			self._workspace_path = os.path.join(os.path.expanduser("~"), APP_NAME)
			if not os.path.isdir(self._workspace_path):
				os.makedirs(self._workspace_path)
			log.info(f"Using default workspace path: {self._workspace_path}")


		try:
			self._run_queue.set_workspace_lock(self._workspace_path) #Creates a lock-file in the folder in question
		except WorkspaceInUseException as exception:
			msgbox = QtWidgets.QMessageBox()
			msgbox.setText(f"<b>Workspace path {self._workspace_path} seems to be in use by another instance of "
		  				f"the {APP_NAME}-app. Do you want to use this workspace path anyway?</b>")
			msgbox.setInformativeText("Continuing might overwrite files from the other instance. "
				"If any other instances are running, please close them first, or change the workspace path. "
				"This issue could also be caused by a previous instance of the app crashing."
				)
			msgbox.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
			msgbox.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
			ret = msgbox.exec()
			if ret == QtWidgets.QMessageBox.StandardButton.No:
				raise WorkspaceInUseException(f"Workspace path {self._workspace_path} is in use by another instance of the "
					f"{APP_NAME}-app. Please close the other instance, change the workspace path, or allow overwriting.") \
						from exception
			else:
				log.warning("User chose to use workspace path anyway, ignoring lock-file.")

		# #Create a lock-file to indicate that the workspace path is in use
		# with open(os.path.join(self._workspace_path, WORKSPACE_LOCK_FILE_NAME), "w", encoding="utf-8") as lock_file:
		# 	lock_file.write("This file is used to indicate that the workspace at this path is in use by another instance of "
		# 		f"the {APP_NAME}-app. Please only remove this file if the app crashed and this file remained.")


		config_save_path = os.path.join(self._workspace_path, "configurations")
		if not os.path.isdir(config_save_path):
			os.makedirs(config_save_path)
			log.info(f"Config output path {config_save_path} is not a valid path, using default path")
		else:
			os.makedirs(config_save_path, exist_ok=True) #Make sure the path exists

		self._config_save_path = config_save_path
		self._config_file_picker_model = FileExplorerModel(
			allow_select_files_only=True
		)


		#========================= load settings =========================
		if settings_in_workspace_path:
			settings_path = os.path.join(self._workspace_path, "settings.ini")
			self._settings = QtCore.QSettings(
				settings_path, QtCore.QSettings.Format.IniFormat, )
			log.info(f"Attempted loading app-settings from {settings_path}. "
	    		f"This resulted in a settings-file loaded from {self._settings.fileName()}")
		else: #Else save
			log.info("Loading settings from default qt-location")
			self._settings = QtCore.QSettings(APP_NAME)
		self._font_point_size = int(
			self._settings.value("font_size", 0, type=int) #type: ignore #If zero->set to system default
		)
		self.set_font_point_size(self._font_point_size)
		self.window.restoreGeometry(self._settings.value(
			"window_geometry", self.window.saveGeometry(), type=QtCore.QRect)) # type: ignore
		new_window_state = self._settings.value("window_state", self.window.windowState())
		if new_window_state != QtCore.Qt.WindowState.WindowNoState:
			self.window.restoreState(new_window_state) # type: ignore


		self._cur_file_path : None | str = self._settings.value("loaded_file_path", None)#The current config. #type: ignore
		assert isinstance(self._cur_file_path, (str, type(None))), (f"Loaded file path should be a string or None, this "
			f"but is a {type(self._cur_file_path)}.")



		#====================== Suboptions window and automatic updating ===================
		self._configuration_model = configuration_model #The configuration model manages creation/changes in the currently
			# loaded configuration

		self._mdi_area = self.ui.ConfigurationMdiArea
		self._cur_option_proxy_models : typing.Dict[str, QtCore.QSortFilterProxyModel]= {}
		self._cur_option_mdi_windows : typing.Dict[str, QtWidgets.QMdiSubWindow] = {}
		self._cur_option_tree_view : typing.Dict[str, DataClassTreeView] = {}
		self._cur_edited_signals : typing.Dict[str, typing.Callable] = {}

		self._configuration_model.proxyModelDictChanged.connect(self.option_proxy_models_changed)
		self.option_proxy_models_changed(self._configuration_model.get_proxy_model_dict()) #Initialize the mdi windows


		#====================== File explorer ===================
		self._config_file_picker_model.setNameFilters(["*.json", "*.yaml", ""])
		self._config_file_picker_model.setNameFilterDisables(False)

		self._config_file_picker_model.setReadOnly(False)
		log.debug(f"Root path used for saving machine learning settings: {self._config_save_path}")
		self._config_file_picker_model.setRootPath(QtCore.QDir.rootPath()) #Subscribe to changes in this path
		self.ui.ConfigFilePickerView.setModel(self._config_file_picker_model)
		self.ui.ConfigFilePickerView.setRootIndex(
			self._config_file_picker_model.index(self._config_save_path)
		)
		self.ui.ConfigFilePickerView.set_double_click_callback(self.file_explorer_item_double_click)
		self.ui.ConfigFilePickerView.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents) #type: ignore
		#Filter only .json and folders


		#======== Open a window which shows the undo/redo stack ========
		if self._configuration_model.undo_stack:
			self.ui.undoView.setStack(self._configuration_model.undo_stack)

		#Un undo-stack change, change the title of the window with a * to indicate that the file has been changed
		if self._configuration_model.undo_stack:
			self._configuration_model.undo_stack.cleanChanged.connect(
				lambda: self.window.setWindowModified(not self._configuration_model.undo_stack.isClean()) #type:ignore
			)

		#Link close-event to a confirmation box
		self.window.closeEvent = self.close_event


		# self._highlight_changed_signal = self._config_file_picker_model.highlightPathChanged.connect(
		# 	self._config_file_picker_model_highlight_path_changed)

		#============= Post-load settings =============
		if self._cur_file_path is not None:
			load_succesfully = self.load_from_file(self._cur_file_path, show_dialog_on_problem=False) #Attempt load
			if not load_succesfully:
				self.new_configuration(ignore_modified_window=True) #If loading failed -> reset to default
		else:
			self.new_configuration(ignore_modified_window=True) #If no file is loaded -> reset to default
		self.ui.ConfigurationMdiArea.tileSubWindows() #Tile the mdi windows

		#==================Console ================
		self._console_item_model = RunQueueConsoleModel()
		self._console_item_model.set_run_queue(self._run_queue)
		self.ui.consoleWidget.set_model(self._console_item_model)

		#Connect right click on console-widget to a context menu
		self.ui.consoleWidget.ui.fileSelectionTableView.setContextMenuPolicy(
			QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
		self.ui.consoleWidget.ui.fileSelectionTableView.customContextMenuRequested.connect(
			self.create_console_context_menu
		)


		#======================== Actions/Shortcuts ========================

		###### Configuration Actions ########
		self.ui.actionNewConfig.triggered.connect(self.new_configuration) #Set to empty config
		self.ui.actionOpenConfig.triggered.connect(self.load_config_from_file_popup)

		self.ui.actionUndo.triggered.connect(self.undo_triggered)
		self.ui.actionRedo.triggered.connect(self.redo_triggered)
		self.ui.actionSave.triggered.connect(self.save_config_triggered)
		self.ui.actionSave_As.triggered.connect(self.save_config_as_triggered)

		######### View Actions ##########
		self.ui.actionIncreaseFontSize.triggered.connect(
			lambda *_: self.set_font_point_size(min(96, self._font_point_size + 1)) #type: ignore
		) #This should be more than enough
		self.ui.actionDecreaseFontSize.triggered.connect(
			lambda *_: self.set_font_point_size(max(3, self._font_point_size - 1)) #type: ignore
		)
		self.ui.actionDefaultFontSize.triggered.connect(
			lambda *_: self.set_font_point_size(0) #0 makes it so that the system default is used
		)
		self.ui.menuMDI_Area.removeAction(self.ui.actionNone) #Remove the "none" action from the mdi menu
		self.ui.ConfigurationMdiArea.add_actions_to_menu(self.ui.menuMDI_Area)

		###### File overview buttons #########
		self.ui.saveCurrentConfigAsBtn.clicked.connect(self.save_config_as_triggered)
		self.ui.saveCurrentConfigBtn.clicked.connect(self.save_config_triggered)
		self.ui.OpenFileLocationBtn.clicked.connect(self.open_save_location_in_explorer)


		###### Run Queue Actions ########
		# self.ui.menuRun_Queue.mousePressEvent().triggered.connect(self.ui.runQueueWidget.show)
		self.ui.menuRun_Queue.aboutToShow.connect(self.expand_run_queue_view_filter_options)
		self.ui.actionBackupRunQueue.triggered.connect(lambda *_: self.ui.runQueueWidget.save_to_file_popup())
		self.ui.actionLoadRunQueue.triggered.connect(lambda *_: self.ui.runQueueWidget.load_from_file_popup())

		###### Buttom-config editor buttons ########
		self.ui.addToQueueButton.clicked.connect(self.add_to_queue_triggered)
		self.ui.saveToQueueItemBtn.clicked.connect(self.save_to_queue_item_triggered)


		#======================================== POST-init ========================================
		self.initial_run_queue_load() #If all went well, try to load existing run queue data


	@catch_show_exception_in_popup_decorator
	def load_config_from_file_popup(self):
		"""Creates a popup that asks the user to select a path, if a path is selected, we will attempt to load the
		configuration from the selected path
		"""
		file_path = QtWidgets.QFileDialog.getOpenFileName(
			self.window, "Select configuration file", self._config_save_path, "Config files (*.json *.yaml)"
		)[0]
		if file_path == "" or file_path is None:
			return
		self.load_from_file(file_path)

	@catch_show_exception_in_popup_decorator
	def save_to_queue_item_triggered(self) -> None:
		"""
		Saves the current configuration to the queue item with the id self._queue_source_id. If it does not exists,
		a keyerror is raised.
		"""
		assert self._cur_source == OptionsSource.QUEUE, (f"Cannot save the config to an existing queue item "
			f"since the current source is {self._cur_source}, not {OptionsSource.QUEUE}")

		try:
			self._run_queue.set_item_config(self._queue_source_id, self._configuration_model.get_configuration_data_copy())
			if self._configuration_model.undo_stack: #Undo stack is saved
				self._configuration_model.undo_stack.setClean()
		except ConfigurationIsFirmException as exception:
			QtWidgets.QMessageBox.warning(self.window, "Cannot save to Queue item", str(exception))
			return
		except KeyError as exception:
			QtWidgets.QMessageBox.warning(self.window, "Cannot save to Queue item",
				f"{exception}. <br> This is likely because the queue item was removed from the queue. <br>"
			)

	def expand_run_queue_view_filter_options(self):
		"""Popules the run queue view filter menu with a copy of the menu provided in the run queue widget"""
		self._cur_run_queue_view_menu = self.ui.runQueueWidget.get_queue_settings_context_menu() #NOTE: even though #pylint: disable=attribute-defined-outside-init
			# we don't use this after creation of the menu, we need to keep a reference for it not to be garbage collected

		for action in self.ui.menuRun_Queue.actions(): #Remove all actions from the menu
			self.ui.menuRun_Queue.removeAction(action)
		for action in self._cur_run_queue_view_menu.actions():
			self.ui.menuRun_Queue.addAction(action)


	def option_proxy_models_changed(self, dict_of_models : typing.OrderedDict[str, QtCore.QSortFilterProxyModel]) -> None:
		"""Upon change of (one of) the dataclass editors -> update the console"""
		#================= Add mdi windows that are new to the config model =================
		for option_name, option_model in dict_of_models.items():
			if option_name in self._cur_option_proxy_models:
				#check if the same model is used
				if self._cur_option_proxy_models[option_name] == option_model:
					log.debug(f"Model for {option_name} has not changed, not updating its model or window")
				else:
					self._cur_option_proxy_models[option_name] = option_model
					self._cur_option_tree_view[option_name].setModel(option_model)
			else: #Create a new mdi window
				log.debug(f"Adding new mdi window for option {option_name}")
				self._cur_option_mdi_windows[option_name] = FramelessMdiWindow()
				self._cur_option_proxy_models[option_name] = option_model
				self._cur_option_tree_view[option_name] = DataClassTreeView()
				self._cur_option_tree_view[option_name].setModel(option_model)
				self._cur_option_tree_view[option_name].setItemDelegate(DataclassEditorsDelegate())#Set custom delegate
				self._cur_option_mdi_windows[option_name].setWidget(self._cur_option_tree_view[option_name])
				self._mdi_area.addSubWindow(self._cur_option_mdi_windows[option_name])
				self._cur_option_mdi_windows[option_name].setWindowTitle(option_name.title().replace("_", " "))
				# self._cur_option_mdi_windows[option_name].show()

		#================= Remove mdi windows that are no longer in the config model =================
		del_windows = [key for key in self._cur_option_mdi_windows if key not in dict_of_models.keys()]
		for window in del_windows:
			self._cur_option_mdi_windows[window].close()
			self._mdi_area.removeSubWindow(self._cur_option_mdi_windows[window])
			del self._cur_option_tree_view[window]
			del self._cur_option_proxy_models[window]
			del self._cur_option_mdi_windows[window]

		self._mdi_area.order_windows_by_windowlist(
			[self._cur_option_mdi_windows[option_name] for option_name in dict_of_models.keys()]
		) #Sort windows according to returned dictionary order

	def create_console_context_menu(self, pos):
		"""Create a context menu at the provided position, displaying all ignored ids. When clicked, un-ignore the id
		and reload the model.

		Args:
			pos (): The local position of the mouse click (sent from self.ui.ConsoleWidget.ui.fileSelectionTableView)
		"""
		ignored_ids = self._console_item_model.get_ignored_ids()

		#Create a context menu with all ids, when clicked, un-ignore the id
		menu = QtWidgets.QMenu()
		#Add text "Re-show ignored ids" to the menu, followed by a splitter

		if len(ignored_ids) == 0:
			action = menu.addAction("(No ignored ids)")
			action.setEnabled(False)
		for cur_id in ignored_ids:
			log.info(f"Ignored ids {ignored_ids} - currently adding item {cur_id} to menu")
			action = menu.addAction(f"(Re)show item: {cur_id}")
			action.triggered.connect(lambda *_, id=cur_id: self._console_item_model.un_ignore_id(id))
		menu.exec(self.ui.consoleWidget.ui.fileSelectionTableView.mapToGlobal(pos))

	def set_run_queue(self, run_queue : RunQueue):
		"""Sets a new runqueue. Indicates that all models need to reload their data.

		Args:
			run_queue (RunQueue): the new runQueue
		"""
		self._run_queue = run_queue
		self.run_queue_table_model : RunQueueTableModel = RunQueueTableModel(self._run_queue)

	@catch_show_exception_in_popup_decorator
	def run_queue_widget_item_double_click(self, index : QtCore.QModelIndex) -> None:
		"""
		Replaces the default behaviour of the on-double click signal of the RunQueueTableView.
		In this method, we first try to load the item, to the configuration-editor.
		If it fails, we just ignore the request. Otherwise mark the item as selected in the model and load
		"""
		cur_id = index.data(RunQueueTableModel.CustomDataRoles.IDRole)
		log.debug(f"Double clicked on index {index.row()} with id {cur_id}")

		ret = self.load_config_from_queue(cur_id)
		if not ret:
			log.warning(f"Failed to load queue item with id {cur_id}, ignoring request")
			return
		else:
			self.run_queue_table_model.set_highligh_by_id(cur_id)

		self.set_source(OptionsSource.QUEUE) #Source from queue
		self._queue_source_id = cur_id #Set the id of the queue item which is currently selected

		log.info(f"Loaded Configuration of queue item with id {cur_id} from RunQueue")

	@catch_show_exception_in_popup_decorator
	def file_explorer_item_double_click(self, index : QtCore.QModelIndex | QtCore.QPersistentModelIndex) -> None:
		"""
		Replaces the default behaviour of the on-double click signal of the FileExplorerView.
		In this method, we first try to load the item, to the configuration-editor.
		If it fails, we just ignore the request. Otherwise mark the item as selected in the model and load it.
		"""
		file_path = self._config_file_picker_model.filePath(index)
		log.debug(f"Double clicked on index {index.row()} with path {file_path}")

		ret = self.load_from_file(file_path)
		if not ret:
			log.warning(f"Failed to load file {file_path}, ignoring request")
			return
		else:
			self._config_file_picker_model.set_highlight_using_index(index)

	@catch_show_exception_in_popup_decorator
	def load_config_from_queue(self, queue_item_id : int) -> bool:
		"""
		Sets the current configuration to the configuration of the queue item with the provided id. If the queue item
		does not exist, raises a KeyError. If queue-item had no valid Configuration-instance, raises a ValueError.

		Args:
			queue_item_id (int): The id of the config we should attemt to load from the current RunQueue
		"""
		#Check if undo stack is clean, if not -> ask user if they want to save
		if self._configuration_model.undo_stack and not self._configuration_model.undo_stack.isClean():
			if not self._ask_overwrite_unsaved_changes():
				log.info("User chose not to overwrite unsaved changes, ignoring request to load config from queue")
				return False

		if queue_item_id == -1:
			log.warning("Queue item id is -1, ignoring request to set current config to queue item")
			raise KeyError("Queue item id is -1, ignoring request to set current config to queue item")

		new_config = self._run_queue.get_item_config(queue_item_id)
		if new_config is None or not isinstance(new_config, Configuration):
			raise ValueError(f"Queue item with id {queue_item_id} did not return a Configuration. Instead, returned "
		    	f"{type(new_config)}.")

		self._configuration_model.set_configuration_data(new_config, validate_after_setting=True)
		if self._configuration_model.undo_stack: #Reset undo stack
			self._configuration_model.undo_stack.clear()

		#========= Update source ===========
		self._queue_source_id = queue_item_id
		self.ui.saveToQueueItemBtn.setHidden(False) #Show the save to queue button
		self._cur_source = OptionsSource.QUEUE

		#========= Update the file picker view ===========
		self._config_file_picker_model.reset_highlight() #Reset the hightlight

		return True

	def set_source(self, new_source : OptionsSource) -> None:
		"""
		Sets the current source of the configuration (self._cur_source).
		If the source changed, the ui will be updated to reflect the new source
		(e.g. show load-to-queue-button, clear any highlights in other source etc.)

		Args:
			new_source (OptionsSource): The new source
		"""
		if new_source == self._cur_source:
			return

		self._cur_source = new_source
		self._cur_file_path = None
		self.ui.saveToQueueItemBtn.setHidden(new_source != OptionsSource.QUEUE) #Show the save button if item from queue

		if new_source == OptionsSource.FILE:
			#Clear highlight from RunQueue
			self.run_queue_table_model.set_highligh_by_id(-1)
		elif new_source == OptionsSource.QUEUE:
			self._config_file_picker_model.reset_highlight() #Reset the hightlight

		# self._config_file_picker_model.reset_highlight()

	def update_ui_by_connection_state(self, new_connection_state : bool) -> None:
		"""Updates the UI based on the connection state (disabled certain buttons, etc).
		Should only be called when runmode is set to network. When running locally - disconnects cannot happen.
		TODO: also put a "reconnect" button in front of the UI?

		Args:
			new_connection_state (bool): The new connection
		"""

		if not new_connection_state:
			self.ui.runQueueOverlayWidget.set_overlay_hidden(False)
			self.ui.ConsoleOverlayWidget.set_overlay_hidden(False)
		else:
			self.ui.runQueueOverlayWidget.set_overlay_hidden(True)
			self.ui.ConsoleOverlayWidget.set_overlay_hidden(True)

	def initial_run_queue_load(self) -> None:
		"""
		Should be called just after setting workspace.
		Tries to load the RunQueue contents from a file according to the filename that is used when the close-event is
		called. If the file does not exist, nothing happens.

		NOTE: In the case of RunQueueClient-based run-mode, this method should probably not be called, as it tries to
			overwrite the run-queue, which might not be desireable in the case of a running server.

		TODO: also move loading settings here?
		"""
		if os.path.exists(os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME)):
			cur_items = self._run_queue.get_all_items_dict_snapshot_copy()
			if len(cur_items.keys()) > 0:
				#Create a popup that asks the user to select a path, if a path is selected, we will attempt to load the
				#Use one-liner
				warning_box = QtWidgets.QMessageBox.question(self.window, "Load run queue?",
					("Trying to reload run-queue from the workspace, but the current run-queue is not empty. Do you want "
					"to load the run-queue from the last session anyway?"),
					QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
				if warning_box == QtWidgets.QMessageBox.StandardButton.No:
					return

			path = os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME)
			self.run_queue_table_model.load_from_file(path, allow_load_running_items="allow")
			# self.ui.runQueueWidget.load_from_file_popup
		else:
			log.info(f"No RunQueue save file found at {os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME)}"
					", continuing with an empty run-queue.")



	def set_font_point_size(self, new_font_size : int) -> None:
		"""
		Set the point size of the font used in this window.
		Args:
			new_font_size (int): The new font size
		"""
		the_window = self.window
		new_font = the_window.font()
		self._font_point_size = new_font_size
		if new_font_size == 0:#Use system default
			self._font_point_size = QtGui.QFont().pointSize()

		new_font.setPointSize(self._font_point_size)
		the_window.setFont(new_font)
		log.info(f"Set font point size to {self._font_point_size}")


	def _save_settings(self) ->None:
		log.info("Saving settings")
		self._settings.setValue("window_geometry", self.window.saveGeometry())
		self._settings.setValue("window_state", self.window.saveState())
		self._settings.setValue("font_size", self._font_point_size)
		self._settings.setValue("loaded_file_path", self._cur_file_path
			  if self._cur_source == OptionsSource.FILE else None) #Only save the path if it was loaded from a file

		for splitter in self.window.findChildren(QtWidgets.QSplitter): #Save the state of all splitters
			self._settings.setValue(f"splitter_state_{splitter.objectName()}", splitter.saveState())

	def open_save_location_in_explorer(self) -> None:
		"""
		Open the folder containing the current config file in the file explorer
		"""
		index = self.ui.ConfigFilePickerView.currentIndex()
		if index.isValid():
			file_path = self._config_file_picker_model.filePath(index)
			if os.path.isfile(file_path):
				file_path = os.path.dirname(file_path)
			QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(file_path))
		else:
			QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(self._config_save_path))

	@staticmethod
	def _ask_overwrite_unsaved_changes() -> bool:
		"""Can be used when the configuration has unsaved changes. Asks the user if they want to overwrite the current
		config with a new one. Returns True if the user wants to overwrite, False otherwise.

		Returns:
			bool: Whether the user wants to overwrite the current config with a new one
		"""
		new_msgbox = QtWidgets.QMessageBox()
		new_msgbox.setWindowTitle("Warning")
		new_msgbox.setIcon(QtWidgets.QMessageBox.Icon.Warning)
		new_msgbox.setText(
			"The current config has unsaved changed, do you want to overwrite them with the selected config?"
		)
		new_msgbox.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
		new_msgbox.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
		new_msgbox.setEscapeButton(QtWidgets.QMessageBox.StandardButton.No)
		ret = new_msgbox.exec()
		if ret == QtWidgets.QMessageBox.StandardButton.Yes:
			return True
		return False

	def new_configuration(self, ignore_modified_window : bool=False) -> None:
		"""
		Create a new configuration with default values. If the current config has unsaved changes, a dialog will be
		shown to ask the user if they want to save the current config first, unless ignore_modified_window is set to
		True.

		Args:
			ignore_modified_window (bool, optional): Whether to ignore the current modified window state. If true,
				overwrites the current config without asking the user. Defaults to False.
		"""
		if self.window.isWindowModified() and not ignore_modified_window:
			if not self._ask_overwrite_unsaved_changes():
				return

			# self._config_file_picker_model.set_highlight_using_path(self._cur_file_path) #Hightlight path -> original
			self.set_source(OptionsSource.FILE)
			self._cur_file_path = None

		self.window.setWindowModified(False)
		self._cur_file_path = None
		self._config_file_picker_model.reset_highlight()
		self._configuration_model.reset_configuration_data_to_default()

	def load_from_file(self,
				new_path : str | None,
				show_dialog_on_problem=True,
				ignore_modified = False
			) -> bool:
		"""Loads the config from a file

		Args:
			new_path (str): The new path from which to load the config
			show_dialog_on_problem (bool, optional): Whether to show a dialog when there are problems. Defaults to True.
			ignore_modified (bool, optional): Whether to ignore the current modified window state. If true,
				overwrites the current config without asking the user. Defaults to False (ask user if modified).

		Returns:
			bool: Whether loading a config was succesful NOTE: still returns True if the passed file was already loaded,
			  also return True if there were problems, returns false if an unhandled exception occurs during loading
		"""
		if new_path is None or not os.path.isfile(new_path):
			return False

		if new_path == self._cur_file_path and self._cur_source == OptionsSource.FILE: #If file is already loaded
			if self._config_file_picker_model.get_highlight_path() != new_path:
				self._config_file_picker_model.set_highlight_using_path(self._cur_file_path) #Do update the highlight path
			return True #Ignore -> but action was successful

		if self.window.isWindowModified() and not ignore_modified:
			msg = QtWidgets.QMessageBox()
			msg.setWindowTitle("Warning")
			msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
			msg.setText("The current config has unsaved changed, do you want to overwrite them with this new config?")
			msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
			msg.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
			msg.setEscapeButton(QtWidgets.QMessageBox.StandardButton.No)
			ret = msg.exec()
			if ret == QtWidgets.QMessageBox.StandardButton.No:
				return False

		self.set_source(OptionsSource.FILE)
		self._cur_file_path = new_path
		self._config_file_picker_model.set_highlight_using_path(self._cur_file_path) #Also update the highlight path
		try:
			if new_path:
				problem_dict = {}
				try: #First try loading using class-type keys that are inside the json file
					problem_dict = self._configuration_model.load_json_from(new_path)
				except (UnkownOptionClassError, NoClassTypesError) as exception:

					msg = QtWidgets.QMessageBox()
					txt = ""
					button_discard = None
					if isinstance(exception, NoClassTypesError):
						txt += ("No class types were found in the file. We can try to to deduce the dataclass-types "
							"by using the option-names and automatic type-deduction. <br><br>")
						txt += ("<br>This could be the "
								"result of a change in the settings format, dataclass module- or class-names, or due to "
								"file corruption.")
					elif isinstance(exception, UnkownOptionClassError):
						txt = "<ul>"
						for option_name, exception_list in exception.args[-1].items():
							txt += f"<b>{option_name}:</b>"
							for exception in exception_list:
								txt += f"<li><b>{type(exception).__name__}</b>:<br>{exception}</li>"
						txt += "</ul>"
						txt += ("<br>There was a problem loading the option dataclass-types from file. This could be the "
								"result of a change in the settings format, dataclass module- or class-names, or due to "
								"file corruption.")
						txt += ("<br><br> We can discard the unknown dataclass-types or try to deduce the "
							"actual dataclass types by using the option-names and automatic type-deduction. <br><br>")

						button_discard = msg.addButton("Discard", QtWidgets.QMessageBox.ButtonRole.NoRole)


					msg.setWindowTitle("Warning")
					msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
					msg.setText(f"Could not immediately load dataclass-types from {new_path}:")
					msg.setInformativeText(txt)

					#TODO: order of buttons is determined by buttonrole, so they seem a bit arbitrary...
					button_deduce = msg.addButton("Deduce", QtWidgets.QMessageBox.ButtonRole.AcceptRole)
					msg.addButton("Cancel", QtWidgets.QMessageBox.ButtonRole.RejectRole)
					msg.setDetailedText(str(exception.args[-1]))


					msg.exec()
					if msg.clickedButton() == button_deduce: #Deduce
						problem_dict = self._configuration_model.load_json_from(new_path, load_using_classtypes_key=False)
					elif msg.clickedButton() == button_discard: #Discard
						problem_dict = self._configuration_model.load_json_from(new_path, ignore_unknown_option_types=True)
					else:
						#Reset settings
						self.new_configuration(ignore_modified_window=True) #If loading failed -> reset to default config
						return False

				if len(problem_dict) > 0 and show_dialog_on_problem:
					msg = QtWidgets.QMessageBox()
					msg.setIcon(QtWidgets.QMessageBox.Icon.Warning) #type: ignore
					msg.setText(
						f"The following problems were encountered while loading the settings from path <a href={new_path}></a>:")
					msg.setWindowTitle("Warning")
					txt =""
					for option_name, exception_list in problem_dict.items():
						txt += f"<b>{option_name}:</b>"
						txt += "<ul>"
						for exception in exception_list:
							txt += f"<li><b>{type(exception).__name__}</b>: {exception}</li>"
						txt += "</ul><br>"

					txt += ("The settings have been loaded anyway, but running them as-is might result in "
						"unexpected behaviour. This could be the result of a change in the settings format, or due to "
						"file corruption.")
					msg.setInformativeText(txt)
					msg.setDetailedText(str(problem_dict))
					msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok) #type: ignore
					msg.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Ok) #type: ignore
					msg.exec()

				try:
					self._configuration_model.validate_current_configuration()
				except OptionTypesMismatch as exception:
					msg = QtWidgets.QMessageBox()
					msg.setWindowTitle("Warning")
					msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
					msg.setText(f"The loaded configuration from <tt>{new_path}</tt> is not stable.")
					replaced_exception = str(exception).replace("\n", "<br>")
					msg.setInformativeText((f"<b>{type(exception).__name__}:</b> {replaced_exception}<br><br>"
			     			"This could be the result of a change in the settings format, the type-deducer, "
							"the option-names, or due to file corruption. <br>"
							"You can continue using the configuration, but any changes to the config will overwrite the "
							" option-subgroup(s) mentioned above.<br>"))
					msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
					msg.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Ok)
					msg.exec()

				self._config_file_picker_model.set_highlight_using_path(new_path)
				return True
		except Exception as exception: #pylint: disable=broad-except #Allow broad-exception, catch all and display
			msg = QtWidgets.QMessageBox()
			msg.setWindowTitle("Error")
			msg.setIcon(QtWidgets.QMessageBox.Icon.Critical) #type: ignore
			msg.setText(f"Could not load config from {new_path}")
			msg.setInformativeText(f"{type(exception).__name__}: {exception}")
			msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok) #type: ignore
			msg.exec()
			log.error(f"Could not load config from {new_path}. Error: {exception}")
			self._cur_file_path = None
			self._config_file_picker_model.reset_highlight()
		return False




	def close_event(self, event : QtGui.QCloseEvent) -> None:
		"""Overload default close event for a confirmation
		"""
		# ConfirmationBox = QtGui.QMessageBox()
		if self.window.isWindowModified():
			quit_msg = "There are unsaved changes. Do you want to save & quit or discard changes before closing?"
			#Create a window with a "quit", "save and quit" and "cancel" button
			win = QtWidgets.QMessageBox()
			win.setWindowTitle("Unsaved changes")
			win.setText(quit_msg)
			win.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Save | QtWidgets.QMessageBox.StandardButton.Discard |
			  QtWidgets.QMessageBox.StandardButton.Cancel)
			win.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Save)
			win.setEscapeButton(QtWidgets.QMessageBox.StandardButton.Cancel)
			win.setIcon(QtWidgets.QMessageBox.Icon.Warning)
			ret = win.exec()


			if ret == QtWidgets.QMessageBox.StandardButton.Save:
				try:
					if not self.save_config_triggered(): #If save was canceled, or failed, stop closing
						event.ignore()
						return
				except Exception as exception: #pylint: disable=broad-except
					log.error(f"Could not save config: {exception}")
					event.ignore() #If save failed - do not close
					return
			elif ret == QtWidgets.QMessageBox.StandardButton.Discard:
				log.info("Discarding changes...")
			else:
				event.ignore()
				return

		if not self.check_if_running_ask_stop_items_before_close(): #Check if user is okay with stopping the queue
			event.ignore()
			return

		save_dict = self._run_queue.get_queue_contents_dict(save_running_as_stopped=True)
		run_queue_contents_path = os.path.join(self._workspace_path, WORKSPACE_RUN_QUEUE_SAVE_NAME)
		with open(run_queue_contents_path, "wb") as file:
			dill.dump(save_dict, file)

		self._save_settings()
		self._run_queue.stop_command_line_queue_emitter() #Stop the command-line queue emitter
		self._run_queue.release_workspace_lock(self._workspace_path) #Release the lock on the workspace
		event.accept()

	def check_if_running_ask_stop_items_before_close(self) -> bool:
		"""Checks if the queue is running an item. If so, asks the user if they want to stop the queue and close anyway.
		Uses a separate function because this function should not cancel the running processes in the case of a
		client-server app.
		"""
		#Check whether queue is running an item - if so, ask for confirmation
		if self._run_queue.get_running_configuration_count() > 0:
			quit_msg = "<b>The queue is currently running an item. Do you want to stop the queue and close anyway?</b>"
			#Create a window with a "quit", "save and quit" and "cancel" button
			win = QtWidgets.QMessageBox()
			win.setWindowTitle("Queue is running")
			win.setText(quit_msg)
			win.setInformativeText("If you choose to stop the queue, all currently running items will be marked as 'stopped'.")
			win.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
			win.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Yes)
			win.setEscapeButton(QtWidgets.QMessageBox.StandardButton.No)
			win.setIcon(QtWidgets.QMessageBox.Icon.Warning)
			ret = win.exec()

			if ret == QtWidgets.QMessageBox.StandardButton.Yes:
				self._run_queue.force_stop_all_running(
					stop_msg="User stopped the RunQueueue when closing the application."
				)
			else:
				return False

		return True



	def get_base_name(self) -> str:
		"""
		Generates a base name for the config file based on the current main-settings
		"""
		if self._cur_file_path:
			# return os.path.basename(self._cur_file_path) #If already saved, use the save-name
			return os.path.basename(self._cur_file_path).rsplit(".", 1)[0] #If already saved, use the save-name

		if hasattr(self._configuration_model, "experiment_name"): #If experiment name is set -> use that
			return self._configuration_model.experiment_name

		# name = ""
		# name += ("_" + self._config_model.data_class) if self._config_model.data_class else ""
		# name += ("_" + self._config_model.task) if self._config_model.task else ""
		# name += ("_" + self._config_model.model) if self._config_model.model else ""
		return "new_config"

	@catch_show_exception_in_popup_decorator
	def add_to_queue_triggered(self):
		"""
		Triggered when the "add to queue" button is clicked. Adds the current config to the queue.
		"""
		log.debug("Add to queue triggered")
		name = self.get_base_name()
		#Ask for user input for the name
		name, ok_clicked = QtWidgets.QInputDialog.getText(self.window, "Add to queue", "Enter a name for the config",
					    QtWidgets.QLineEdit.EchoMode.Normal, name)
		if ok_clicked:
			# self.ml_queue.add_to_queue(name, self._options.get_options_data_copy())
			self.run_queue_table_model.add_to_queue(name, self._configuration_model.get_configuration_data_copy())

	def undo_triggered(self):
		"""
		Triggered when undo is clicked. Undoes the last change to the options
		"""
		log.info("Undo triggered")
		if self._configuration_model.undo_stack:
			self._configuration_model.undo_stack.undo()

	def redo_triggered(self):
		"""
		Triggered when redo is clicked. Redoes the last change to the options
		"""

		log.info("Redo triggered")
		if self._configuration_model.undo_stack:
			self._configuration_model.undo_stack.redo()

	def save_config_triggered(self) -> bool:
		"""Save the config to the current file path. If no file path is set, calls save_config_as_triggered instead.

		Returns:
			bool: Whether the save was successful
		"""
		log.info("Save triggered")
		if self._cur_file_path is None or (self._cur_source != OptionsSource.FILE) or not os.path.isfile(self._cur_file_path):
			ret = self.save_config_as_triggered() #Already cleans undo-stack if successful
		else:
			ret = self._configuration_model.save_json_to(self._cur_file_path)
			if ret and self._configuration_model.undo_stack: #If save was successful -> set clean state
				self._configuration_model.undo_stack.setClean()
				self.window.setWindowModified(False)
		return ret

	def save_config_as_triggered(self) -> bool:
		"""
		Triggered when the "save as" button is clicked. Opens a file dialog and saves the current config to the selected
		Returns:
			bool: True if the save was successful, False otherwise.
		"""
		log.info("Save as triggered")
		ret = False
		start_path = self._config_save_path
		if self._cur_file_path is not None\
				and self._cur_source == OptionsSource.FILE\
				and os.path.isfile(self._cur_file_path): #If already saved -> use the current file as start path
			start_path = self._cur_file_path

		file_path, _ = QtWidgets.QFileDialog.getSaveFileName(
			None, "Save config", start_path, "JSON (*.json)") #type:ignore
		if file_path:
			self._cur_file_path = file_path #If source is file -> this sets current file path
			self._config_file_picker_model.set_highlight_using_path(self._cur_file_path) #Also update the highlighted path
			ret = self._configuration_model.save_json_to(file_path)
			log.info(f"Save as returned {ret}")
			if ret and self._configuration_model.undo_stack: #If save was successful -> set clean state
				self._configuration_model.undo_stack.setClean()
				self.window.setWindowModified(False)
		else:
			log.info("No Save Path Selected")
		return ret
			#TODO: set selection to current file


def run_example_local_app(log_level : int=logging.INFO):
	"""Runs the example app"""
	#pylint: disable=import-outside-toplevel
	from configurun.examples.example_target_function import example_target_function
	from configurun.examples.example_configuration import example_deduce_new_option_classes
	import tempfile
	from configurun.app.create_run import run_local

	tempdir = tempfile.gettempdir()
	workspace_path = os.path.join(tempdir, APP_NAME, "Configurun-Local-Example")
	run_local(
		target_function=example_target_function,
		options_source=example_deduce_new_option_classes,
		workspace_path=workspace_path,
		log_level=log_level
	)

if __name__ == "__main__":
	run_example_local_app(log_level=logging.DEBUG)




"""
Convenience functions to quickly create & run local or networked app-instances (client/local_app) with the specified
target function and option source.
"""
import argparse
import logging
import os
import sys
import typing

from PySide6 import QtCore, QtWidgets

from configurun.classes.run_queue import RunQueue
from configurun.classes.run_queue_client import RunQueueClient
from configurun.configuration.base_options import BaseOptions
from configurun.configuration.configuration import Configuration
from configurun.configuration.configuration_model import ConfigurationModel
from configurun.app.main_window import APP_NAME, MainWindow
from configurun.app.network_main_window import NetworkMainWindow
from configurun.configuration.argparse_to_dataclass import argparse_to_dataclass

log = logging.getLogger(__name__)


def _get_option_function(option_source:\
			typing.Callable[[Configuration], typing.Dict[str, typing.Type[BaseOptions] | typing.Type[None]]] | \
			argparse.ArgumentParser | \
			typing.Type[BaseOptions]
		) -> typing.Callable[[Configuration], typing.Dict[str, typing.Type[BaseOptions] | typing.Type[None]]]:
	"""Get the option-function from the passed argument.
	User is able to pass:
	- A function returning a dict of option (dataclass) objects,
	- A single BaseOptions object
	- An argparse.ArgumentParser object

	In this function, all are converted to the first - a function returning a dict of option objects.
	"""

	if isinstance(option_source, argparse.ArgumentParser):
		# raise NotImplementedError("argparse.ArgumentParser not yet implemented")
		argparse_dataclass = argparse_to_dataclass(option_source, "ArgparseOptions")
		def argparse_option_function(*_): #argparse-options: Return the argparse-dataclass
			return {"Options" : argparse_dataclass}
		return argparse_option_function #type: ignore
	elif isinstance(option_source, type) and issubclass(option_source, BaseOptions):
		def option_function(*_): #'Dumb'-options: Just return the option type
			return {"Options" : option_source}
		return option_function #type: ignore
	elif callable(option_source):
		return option_source
	else:
		raise ValueError(f"Invalid option_source: {option_source} of type {type(option_source)}")




def run_local(
		target_function : typing.Callable,
		options_source : \
			typing.Callable[[Configuration], typing.Dict[str, typing.Type[BaseOptions] | typing.Type[None]]] | \
			argparse.ArgumentParser | \
			typing.Type[BaseOptions]
				,
		workspace_path : str = "",
		create_workspace_path : bool = True, #Whether to create the workspace path if it does not exist
		run_queue_n_processes : int = 1,
		log_level : int = logging.INFO,
		run_queue_kwargs : typing.Optional[typing.Dict[str, typing.Any]] = None,
		config_model_kwargs : typing.Optional[typing.Dict[str, typing.Any]] = None,

	):
	"""Convenience function that creates and runs a local instance of the app with the specified target function and 
	option source. Workspace path is option, if none is provided, uses '~/Configurun/'.

	Args:
		target_function (typing.Callable): The target function on which tasks will be run, this function should take
			a configuration as the first argument, and the rest of the arguments should be the arguments passed to the
			```RunQueue._process_queue_item()```-method.

		options_source (typing.Callable[[Configuration], typing.Dict[str, typing.Type[BaseOptions]
		 		|  typing.Type[None]]]
				|, optional):
			The source of the options in the ui. This can either be:
				- A function returning a dict of option (dataclass) class,
				- A single BaseOptions-type class
				- An argparse.ArgumentParser class
			Defaults to None.
		workspace_path (str, optional): The path to the workspace folder. Attempts to load progress from here, also saves
		 	progress to here. Defaults to "". If empty/default, the default workspace folder is used (~/Configurun/)
		create_workspace_path (bool, optional): Whether to create the workspace path if it does not exist.
			Defaults to True.
		run_queue_n_processes (int, optional): The number of processes to use in the run queue. Defaults to 1.
		log_level (int, optional): The log level to use. Defaults to logging.INFO
		run_queue_kwargs (typing.Dict[str, typing.Any], optional): The keyword arguments passed to the
			RunQueue constructor. Defaults to {}.
			E.g.:
			- log_location (str) : The path where the log file should be outputted (should be a folder)
				if blank, use default location based on workspace folder
		config_model_kargs (typing.Dict[str, typing.Any], optional): The keyword arguments passed to the
			ConfigurationModel constructor. Defaults to {}.
			E.g.:
			- use_cache (bool): Whether to use the cache or not to temporarily save configurations. This makes it so
				option-group settings are remembered when switching back/forth (for example between 2 model-options).
				Defaults to True. If true undo_stack is also used
			- use_undo_stack (bool): Whether to use the undo stack or not. Defaults to True

	"""
	#=========== Initialize logger ===========
	formatter = logging.Formatter("[{pathname:>90s}:{lineno:<4}]  {levelname:<7s}   {message}", style='{')
	handler = logging.StreamHandler()
	handler.setFormatter(formatter)
	logging.basicConfig(
		handlers=[handler],
		level=log_level) #Without time
	root_logger = logging.getLogger()
	root_logger.setLevel(log_level)


	if run_queue_kwargs is None:
		run_queue_kwargs = {}
	if config_model_kwargs is None:
		config_model_kwargs = {}

	app = QtWidgets.QApplication(sys.argv)

	if workspace_path == "" or workspace_path is None: #Set to default workspace path if not set
		workspace_path = os.path.join(os.path.expanduser("~"), APP_NAME)
		log.info(f"No workspace path provided, using default: {workspace_path}")

	if not create_workspace_path and not os.path.exists(workspace_path):
		raise ValueError(f"Workspace path {workspace_path} does not exist and workspace-creation is set to False.")
	os.makedirs(workspace_path, exist_ok=True) #Create the workspace folder if it does not exist yet
	QtCore.QDir.setCurrent(workspace_path) #Set the current working directory to the workspace path

	run_queue = RunQueue(
		target_function=target_function,
		n_processes=run_queue_n_processes,
		log_location=os.path.join(workspace_path, "logs"),
		**run_queue_kwargs
	) #Create the run queue
	options_function = _get_option_function(options_source) #Create the options-function

	#Select a config-model, pass the deduce_new_option_class_types function to the constructor
	config_model = ConfigurationModel(
		option_type_deduction_function=options_function,
		**config_model_kwargs
	)

	#Create the Qt-main window in which the app will be placed
	main_window = QtWidgets.QMainWindow()

	#Create the actual app
	MainWindow(
		configuration_model=config_model,
		run_queue=run_queue,
		window=main_window,
		workspace_path=workspace_path
	)

	main_window.show() #Show the window
	app.exec()

def run_client(
			options_source : \
				typing.Callable[[Configuration], typing.Dict[str, typing.Type[BaseOptions] | typing.Type[None]]] | \
				argparse.ArgumentParser | \
				typing.Type[BaseOptions],
			workspace_path : str = "",
			create_workspace_path : bool = True, #Whether to create the workspace path if it does not exist
			log_level : int = logging.INFO,
			config_model_kwargs : typing.Optional[typing.Dict[str, typing.Any]] = None,
		):

	"""
	Convenience function that creates and runs a client instance of the app with the specified option source. 
	Differs from local_app in that no target_function is provided as this should be provided on the server-side.
	Workspace path is optional, if none is provided, uses '~/Configurun/'.

	Args:
		options_source (typing.Callable[[Configuration], typing.Dict[str, typing.Type[BaseOptions]
				|  typing.Type[None]]]
				|, optional):
			The source of the options in the ui. This can either be:
				- A function returning a dict of option (dataclass) objects,
				- A single BaseOptions object
				- An argparse.ArgumentParser object
			Defaults to None.

		workspace_path (str, optional): The path to the workspace folder. Attempts to load progress from here, also saves
			configs/logs/settings to here. Defaults to "". If empty/default, the default workspace folder is used
			(~/Configurun-Client/)

		create_workspace_path (bool, optional): Whether to create the workspace path if it does not exist.
			Defaults to True.

		config_model_kwargs (typing.Dict[str, typing.Any], optional): The keyword arguments passed to the
			ConfigurationModel constructor. Defaults to {}.
			E.g.:
			- use_cache (bool): Whether to use the cache or not to temporarily save configurations. This makes it so
				option-group settings are remembered when switching back/forth (for example between 2 model-options).
				Defaults to True. If true undo_stack is also used
			- use_undo_stack (bool): Whether to use the undo stack or not. Defaults to True

		log_level (int, optional): The log level to use. Defaults to logging.INFO
	"""
	#=========== Initialize logger ===========
	formatter = logging.Formatter("[{pathname:>90s}:{lineno:<4}]  {levelname:<7s}   {message}", style='{')
	handler = logging.StreamHandler()
	handler.setFormatter(formatter)
	logging.basicConfig(
		handlers=[handler],
		level=log_level) #Without time
	root_logger = logging.getLogger()
	root_logger.setLevel(log_level)
	root_logger.handlers=[handler]


	if config_model_kwargs is None:
		config_model_kwargs = {}

	app = QtWidgets.QApplication(sys.argv)

	if workspace_path == "" or workspace_path is None: #Set to default workspace path if not set
		workspace_path = os.path.join(os.path.expanduser("~"), APP_NAME+'-Client')
		log.info(f"No workspace path provided, using default: {workspace_path}")

	if not create_workspace_path and not os.path.exists(workspace_path):
		raise ValueError(f"Workspace path {workspace_path} does not exist and workspace-creation is set to False.")
	os.makedirs(workspace_path, exist_ok=True) #Create the workspace folder if it does not exist yet
	QtCore.QDir.setCurrent(workspace_path) #Set the current working directory to the workspace path

	options_function = _get_option_function(options_source) #Create the options-function

	#Select a config-model, pass the deduce_new_option_class_types function to the constructor
	config_model = ConfigurationModel(option_type_deduction_function=options_function, **config_model_kwargs)

	#Create the Qt-main window in which the app will be placed
	main_window = QtWidgets.QMainWindow()

	run_queue_client = RunQueueClient(
	)


	#Create the actual app
	NetworkMainWindow(
		configuration_model=config_model,
		run_queue_client=run_queue_client,
		window=main_window,
		workspace_path=workspace_path,
	)

	main_window.show() #Show the window
	app.exec()
	exit(0)



if __name__ == "__main__":
	from configurun.examples.example_configuration import \
	    example_deduce_new_option_classes
	from configurun.examples.example_target_function import example_target_function
	run_local(
		target_function=example_target_function,
		options_source=example_deduce_new_option_classes,
		run_queue_n_processes=1,
		log_level=logging.DEBUG
	)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'network_login_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.5.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFormLayout, QHBoxLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QVBoxLayout, QWidget)

class Ui_NetworkLoginWidget(object):
    def setupUi(self, NetworkLoginWidget):
        if not NetworkLoginWidget.objectName():
            NetworkLoginWidget.setObjectName(u"NetworkLoginWidget")
        NetworkLoginWidget.resize(452, 154)
        self.verticalLayout = QVBoxLayout(NetworkLoginWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.label = QLabel(NetworkLoginWidget)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label)

        self.serverIPComboBox = QComboBox(NetworkLoginWidget)
        self.serverIPComboBox.setObjectName(u"serverIPComboBox")
        self.serverIPComboBox.setEditable(True)

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.serverIPComboBox)

        self.label_2 = QLabel(NetworkLoginWidget)
        self.label_2.setObjectName(u"label_2")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_2)

        self.serverPortComboBox = QComboBox(NetworkLoginWidget)
        self.serverPortComboBox.setObjectName(u"serverPortComboBox")
        self.serverPortComboBox.setEditable(True)

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.serverPortComboBox)

        self.label_3 = QLabel(NetworkLoginWidget)
        self.label_3.setObjectName(u"label_3")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.label_3)

        self.serverPasswordLineEdit = QLineEdit(NetworkLoginWidget)
        self.serverPasswordLineEdit.setObjectName(u"serverPasswordLineEdit")
        self.serverPasswordLineEdit.setEchoMode(QLineEdit.Normal)

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.serverPasswordLineEdit)


        self.verticalLayout.addLayout(self.formLayout)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.cancelBtn = QPushButton(NetworkLoginWidget)
        self.cancelBtn.setObjectName(u"cancelBtn")

        self.horizontalLayout.addWidget(self.cancelBtn)

        self.disconnectBtn = QPushButton(NetworkLoginWidget)
        self.disconnectBtn.setObjectName(u"disconnectBtn")

        self.horizontalLayout.addWidget(self.disconnectBtn)

        self.connectBtn = QPushButton(NetworkLoginWidget)
        self.connectBtn.setObjectName(u"connectBtn")

        self.horizontalLayout.addWidget(self.connectBtn)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.verticalLayout.setStretch(0, 1)

        self.retranslateUi(NetworkLoginWidget)

        QMetaObject.connectSlotsByName(NetworkLoginWidget)
    # setupUi

    def retranslateUi(self, NetworkLoginWidget):
        NetworkLoginWidget.setWindowTitle(QCoreApplication.translate("NetworkLoginWidget", u"Form", None))
        self.label.setText(QCoreApplication.translate("NetworkLoginWidget", u"Server IP", None))
        self.label_2.setText(QCoreApplication.translate("NetworkLoginWidget", u"Server Port", None))
        self.label_3.setText(QCoreApplication.translate("NetworkLoginWidget", u"Password", None))
        self.cancelBtn.setText(QCoreApplication.translate("NetworkLoginWidget", u"Cancel", None))
        self.disconnectBtn.setText(QCoreApplication.translate("NetworkLoginWidget", u"Disconnect", None))
        self.connectBtn.setText(QCoreApplication.translate("NetworkLoginWidget", u"Connect", None))
    # retranslateUi




"""
Declaration separate from local_app and client_app to enable the server to skip importing PySide6 altogether.
"""
import logging
import os
import typing
import time

from configurun.classes.run_queue import RunQueue
from configurun.classes.run_queue_server import RunQueueServer
log = logging.getLogger(__name__)

def _cleanup_server(run_queue_server : RunQueueServer):
	"""
	Callback function to, on process end (Ctrl+C), cleanup the server and close the app

	Args:
		app (QtCore.QCoreApplication): The app in which the server is running
		run_queue_server (RunQueueServer): The runqueue-server instance
	"""
	log.info("Received shutdown signal, now attempting to cleaning up server and close app")
	run_queue_server.terminate() #Disconnect all, stop running and save progress
	# app.quit()

	log.info("Server cleanup complete, exiting")

def run_server(
			target_function : typing.Callable,
			password : str,
			workspace_path : str = "",
			run_queue_n_processes : int = 1,
			hostname : str = "localhost",
			port : int = 5454,
			log_level : int = logging.INFO,
			run_queue_kwargs : typing.Optional[typing.Dict[str, typing.Any]] = None
		):
	"""
	WARNING: RUNNING A SERVER ALLOWS OTHER MACHINES ON THIS NETWORK TO EXECUTE ARBITRARY CODE IF THEY KNOW THE PASSWORD
	PLEASE RUN THIS IN A TRUSTED NETWORK ENVIRONEMENT. Run at your own risk.

	Convenience function that constructs a local instance of the runqueue-server with the specified target function.
	It then runs the server in a QtCore app.

	Args:
		target_function (typing.Callable): The target function on which tasks will be run, this function should take
			a configuration as the first argument, and the rest of the arguments should be the arguments passed to the
			```RunQueue._process_queue_item()```-method.

		workspace_path (str, optional): The path to the workspace folder. Attempts to load progress from here, also saves
			progress to here. Defaults to "". If empty/default, the default workspace folder is used (~/Configurun-server/)

		run_queue_n_processes (int, optional): The number of processes to use in the run queue. Defaults to 1.

		run_queue_kwargs (typing.Dict[str, typing.Any], optional): The keyword arguments passed to the
			RunQueue constructor. Defaults to {}.
			Possible kwargs:
				log_location (str) : The path where the log file should be outputted (should be a folder)
					if blank, use default location using TEMP folder
	"""
	#=========== Initialize logger ===========
	formatter = logging.Formatter("[{pathname:>90s}:{lineno:<4}]  {levelname:<7s}   {message}", style='{')
	handler = logging.StreamHandler()
	handler.setFormatter(formatter)
	logging.basicConfig(
		handlers=[handler],
		level=log_level) #Without time
	root_logger = logging.getLogger()
	root_logger.setLevel(log_level)



	if run_queue_kwargs is None:
		run_queue_kwargs = {}

	assert len(password) > 0, "Password for server cannot be empty"

	if workspace_path == "" or workspace_path is None:
		workspace_path = os.path.join(os.path.expanduser("~"), "Configurun-Server")
		log.info(f"No workspace path provided, using default: {workspace_path}")

	os.makedirs(workspace_path, exist_ok=True) #Create the workspace folder if it does not exist yet

	runqueue = RunQueue(
		target_function=target_function,
		n_processes=run_queue_n_processes,
		log_location=os.path.join(workspace_path, "logs"),
		**run_queue_kwargs
	)

	run_queue_server = None
	try:
		run_queue_server = RunQueueServer(
				run_queue=runqueue,
				password=password,
				hostname=hostname,
				port=port,
				workspace_path=workspace_path
		)
		run_queue_server.run()
		log.info("Server started, listening for client connections...")
		while True:
				time.sleep(1)

	except KeyboardInterrupt:
		log.info("Received keyboard interrupt, now attempting to cleanup server and close app...")
	except Exception as exception: #pylint: disable=broad-except
		log.exception(f"{type(exception).__name__}: {exception}")
		log.info("Due to the above unhandled exception, now attempting to cleanup server and close app...")
	finally:
		if run_queue_server:
			_cleanup_server(run_queue_server)
		exit(0)



"""The setup script."""
from setuptools import setup, find_packages


setup(
	name = "configurun",
	version= "0.2.5",
	packages=find_packages('.'),
    description=("Cross-platform user-interface tools to create and manage (machine learning) configurations and run them automatically and remotely."),
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author="Wouter Stokman",
    url="https://github.com/Woutah/configurun",
    license="LGPLv2",
    include_package_data=True,
    install_requires=[ #Generated using pipreqs
        'PySide6>=6.0.0', # Qt for Python, works for 6.5.1.1
        'pathos>=0.3.0', #Works for 0.3.0
        'setuptools>=65.0.0', #Works for 65.5.0
		'dill>=0.3.0', #Works for 0.3.6
		'multiprocess>=0.70.00', #Works for 0.70.14
		'numpydoc>=1.4.0', #Works for 1.5.0
		'pycryptodome>=3.10.0', #Works for 3.18.0
        'pyside6-utils>=1.2.1, <1.3.0', #Works for 1.2.1
        'PySignal>=1.1.1' #Works for 1.1.1,
	]
)
